# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import inspect
sys.path.insert(0, os.path.abspath('.'))

__location__ = os.path.join(os.getcwd(), os.path.dirname(
    inspect.getfile(inspect.currentframe())))

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# add 'ray-optics/src' to path
sys.path.insert(0, os.path.join(__location__, '../../src'))

# -- Project information -----------------------------------------------------

project = u'ray-optics'
copyright = u'2017-2020, Michael J. Hayford'
author = u'Michael J. Hayford'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ''
# The full version, including alpha/beta/rc tags.
release = ''

try:
    from rayoptics import __version__ as version
except ImportError:
    pass
else:
    release = version

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.intersphinx', 'sphinx.ext.todo',
              'sphinx.ext.autosummary', 'sphinx.ext.viewcode', 'sphinx.ext.coverage',
              'sphinx.ext.doctest', 'sphinx.ext.ifconfig', 'sphinx.ext.mathjax',
              'sphinx.ext.napoleon']

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['Thumbs.db', '.DS_Store', 'tests']

# The master toctree document.
master_doc = 'index'

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

rst_prolog = """
.. |ybar| replace:: :math:`y-\overline{y}`
.. |nubar| replace:: :math:`\omega-\overline{\omega}`
.. |minimum_python_version| replace:: 3.7
.. |minimum_numpy_version| replace:: 1.15.0
.. |Series| replace:: :class:`pandas.Series`
.. |DataFrame| replace:: :class:`pandas.DataFrame`
"""

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['rayoptics.']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#html_theme = 'alabaster'
html_theme = 'sphinx_rtd_theme'


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- External mapping ------------------------------------------------------------
python_version = '.'.join(map(str, sys.version_info[0:2]))
intersphinx_mapping = {
    'sphinx': ('https://www.sphinx-doc.org/en/stable', None),
    'python': ('https://docs.python.org/' + python_version, None),
    'matplotlib': ('https://matplotlib.org', None),
    'numpy': ('https://docs.scipy.org/doc/numpy', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference', None),
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2019 Michael J. Hayford
""" module to facilitate first order definition of an optical model

.. Created on Thu May 16 19:57:47 2019

.. codeauthor: Michael J. Hayford
"""
import math

from rayoptics.optical import firstorder
from rayoptics.optical.idealimager import IdealImager, ideal_imager_setup

from rayoptics.util import dict2d
from rayoptics.util.dict2d import dict2D

from rayoptics.optical import etendue
from rayoptics.optical.etendue import (obj_img_set, fld_ape_set,
                                       fld_labels, ap_labels)


def create_specsheets():
    specsheets = {}

    # setup finite conjugate defaults
    fev = dict2D(fld_ape_set, obj_img_set)
    fev['field']['object'] = dict([(fld_labels[0], None)])
    fev['aperture']['object'] = dict([(ap_labels[2], None),
                                      (ap_labels[1], None)])
    fev['field']['image'] = dict([(fld_labels[0], None)])
    fev['aperture']['image'] = dict([(ap_labels[2], None),
                                     (ap_labels[1], None)])
    fss = SpecSheet('finite', etendue_values=fev)
    specsheets['finite'] = fss

    # setup infinite conjugate defaults
    imager_inputs = {'s': -math.inf}
    imager = IdealImager(None, -math.inf, None, None, None)

    iev = dict2D(fld_ape_set, obj_img_set)
    iev['field']['object'] = dict([(fld_labels[1], None)])
    iev['aperture']['object'] = dict([(ap_labels[0], None)])
    iev['field']['image'] = dict([(fld_labels[0], None)])
    iev['aperture']['image'] = dict([(ap_labels[2], None),
                                     (ap_labels[1], None)])

    ifss = SpecSheet('infinite', imager=imager,
                     imager_inputs=imager_inputs,
                     frozen_imager_inputs=[True, True, True, True, False],
                     etendue_values=iev)
    specsheets['infinite'] = ifss

    return specsheets


def create_specsheet_from_model(opt_model, specsheets=None):
    if specsheets is None:
        specsheets = create_specsheets()

    specsheet = opt_model.specsheet
    if specsheet is None:
        conj_type = 'finite'
        if opt_model.seq_model.gaps[0].thi > 10e8:
            conj_type = 'infinite'
        specsheet = specsheets[conj_type]
    firstorder.specsheet_from_parax_data(opt_model, specsheet)
    opt_model.specsheet = specsheet
    return specsheet


class SpecSheet():
    """ First order optical specification for :class:`~.OpticalModel`

    Attributes:
        conjugate_type: one of `infinite`, `finite`
        imager: instance of `IdealImager`
        imager_inputs: dict of inputs to `ideal_imager_setup`
        frozen_imager_inputs: list of booleans, if True the parameter is frozen
        etendue_inputs: field and aperture inputs used to define the etendue
        etendue_values: dict2D of aperture/field vs object/image
        partitions: 'imager', 'field', and 'aperture'; number of items in each
    """
    def __init__(self, conjugate_type,
                 imager=None, imager_inputs=None, frozen_imager_inputs=None,
                 etendue_inputs=None, etendue_values=None):
        self.conjugate_type = conjugate_type

        if imager is None:
            imager = IdealImager(None, None, None, None, None)
        self.imager = imager

        self.imager_inputs = imager_inputs if imager_inputs else {}

        self.frozen_imager_inputs = (frozen_imager_inputs
                                     if frozen_imager_inputs
                                     else [False]*5)

        self.etendue_inputs = (etendue_inputs if etendue_inputs
                               else dict2D(fld_ape_set, obj_img_set))
        self.etendue_values = (etendue_values if etendue_values
                               else dict2D(fld_ape_set, obj_img_set))

        self.partition_defined()

    def __json_encode__(self):
        attrs = dict(vars(self))
        if hasattr(self, 'partitions'):
            del attrs['partitions']
        return attrs

    def sync_to_restore(self, opt_model):
        # imager is exported as a list. convert back to an IdealImager
        self.imager = IdealImager(*self.imager)

    def imager_defined(self):
        """ compute imager and etendue values given input dicts """
        if self.conjugate_type == 'finite':
            imager_defined = 'm' if self.imager.m is not None else False
        else:
            imager_defined = 'f' if self.imager.f is not None else False
        return imager_defined

    def partition_defined(self):
        """ which partition defines the imager or None """
        num_imager_inputs = len(self.imager_inputs)
        li = dict2d.num_items_by_type(self.etendue_inputs,
                                      fld_ape_set, obj_img_set)
        num_field_inputs = li['field']
        num_aperture_inputs = li['aperture']
        partitions = {'imager': num_imager_inputs,
                      'field': num_field_inputs,
                      'aperture': num_aperture_inputs}
        self.partitions = partitions
        max_partition = max(partitions, key=partitions.get)
        max_num_inputs = partitions[max_partition]
        return max_partition if max_num_inputs == 2 else None, max_num_inputs

    def generate_from_inputs(self, imgr_inputs, etendue_inputs):
        """ compute imager and etendue values given input dicts """
        max_partition, max_num_inputs = self.partition_defined()
        num_imager_inputs = self.partitions['imager']
        num_field_inputs = self.partitions['field']
        num_aperture_inputs = self.partitions['aperture']

        conj_type = self.conjugate_type

        imager_inputs = {}
        if max_num_inputs <= 1:
            # fill in imager_inputs with any previous calculations for m or f
            if conj_type == 'finite':
                if num_imager_inputs < 2 and self.imager.m is not None:
                    imager_inputs['m'] = self.imager.m
            else:
                if num_imager_inputs < 2 and self.imager.f is not None:
                    imager_inputs['f'] = self.imager.f

        # update imager_inputs with user entries
        imager_inputs.update(imgr_inputs)
        imager_inputs = {k: v for (k, v) in imager_inputs.items()
                         if v is not None}

        # calculate an ideal imager for imager_inputs
        imager = ideal_imager_setup(**imager_inputs)

        if conj_type == 'finite':
            imager_defined = True if imager.m is not None else False
        else:
            imager_defined = True if imager.f is not None else False

        etendue_values = self.etendue_values

        for fa_key, fa_value in etendue_inputs.items():
            for oi_key, oi_value in fa_value.items():
                etendue.fill_in_etendue_data(imager, fa_key,
                                             etendue_inputs[fa_key][oi_key],
                                             etendue_values[fa_key][oi_key])

        if imager_defined:
            if num_field_inputs >= 1 and num_aperture_inputs >= 1:
                # we have enough data to calculate all of the etendue grid
                ii = etendue.do_etendue_via_imager(conj_type, imager,
                                                   etendue_inputs,
                                                   etendue_values)

                if ii:
                    imager_inputs[ii[0]] = ii[1]
                    imager = ideal_imager_setup(**imager_inputs)
                    etendue.do_etendue_via_imager(conj_type, imager,
                                                  etendue_inputs,
                                                  etendue_values)
            elif num_field_inputs == 1:
                # we have enough data to calculate all of the etendue grid
                row = dict2d.row(etendue_inputs, 'field')
                obj_img_key = 'object' if len(row['object']) else 'image'
                etendue.do_field_via_imager(conj_type, imager, etendue_inputs,
                                            obj_img_key, etendue_values)
            elif num_aperture_inputs == 1:
                # we have enough data to calculate all of the etendue grid
                row = dict2d.row(etendue_inputs, 'aperture')
                obj_img_key = 'object' if len(row['object']) else 'image'
                etendue.do_aperture_via_imager(conj_type, imager,
                                               etendue_inputs, obj_img_key,
                                               etendue_values)
        else:  # imager not specified
            if num_field_inputs == 2 or num_aperture_inputs == 2:
                fld_ape_key = 'field' if num_field_inputs == 2 else 'aperture'
                # solve for imager
                ii = etendue.do_etendue_to_imager(fld_ape_key, etendue_inputs,
                                                  etendue_values)
                imager_inputs[ii[0]] = ii[1]
                imager = ideal_imager_setup(**imager_inputs)
                # update etendue grid
                etendue.do_etendue_via_imager(conj_type, imager,
                                              etendue_inputs, etendue_values)

        self.imager = imager
        self.etendue_values = etendue_values
        return imager, etendue_values



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2019 Michael J. Hayford
""" script file providing an environment for using rayoptics

.. Created on Sun Feb 10 22:42:36 2019

.. codeauthor: Michael J. Hayford
"""

# initialization
import math
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from pathlib import Path
import pandas as pd

# import ipywidgets as widgets
# from ipywidgets import interact, interactive, fixed, interact_manual
# from IPython.display import display

# ray-optics
import rayoptics
from rayoptics.gui.appmanager import AppManager, ModelInfo
from rayoptics.gui.appcmds import create_new_model, open_model

# optical model
from rayoptics.optical.opticalmodel import OpticalModel

import rayoptics.optical.surface as srf
from rayoptics.optical.gap import Gap
from rayoptics.optical import elements
from rayoptics.optical.thinlens import ThinLens
from rayoptics.optical.profiles import (Spherical, Conic, EvenPolynomial,
                                        RadialPolynomial)
from rayoptics.optical.opticalspec import (WvlSpec, FieldSpec, Field,
                                           PupilSpec, FocusRange)
from rayoptics.optical.model_enums import (DimensionType, DecenterType)

# ray-optics first and third order
import rayoptics.optical.firstorder as fo
from rayoptics.optical.firstorder import compute_first_order
import rayoptics.optical.thirdorder as to
from rayoptics.optical.thirdorder import compute_third_order

# ray tracing
from rayoptics.optical.trace import (RayPkg, RaySeg, list_ray,
                                     trace, trace_base, trace_with_opd,
                                     trace_astigmatism)
import rayoptics.optical.raytrace as rt

# paraxial design
import rayoptics.optical.model_constants as mc

import rayoptics.mpl.paraxdgnfigure as pdfig
from rayoptics.mpl.paraxdgnfigure import ParaxialDesignFigure
from rayoptics.mpl.interactivediagram import InteractiveDiagram

# axis array figures
from rayoptics.mpl.axisarrayfigure import Fit
from rayoptics.mpl.axisarrayfigure import RayFanFigure
from rayoptics.mpl.axisarrayfigure import SpotDiagramFigure
from rayoptics.mpl.axisarrayfigure import WavefrontFigure

from rayoptics.mpl import analysisplots

# lens layout
import rayoptics.mpl.lenslayoutfigure as lay
from rayoptics.mpl.lenslayoutfigure import LensLayoutFigure
from rayoptics.mpl.interactivelayout import InteractiveLayout



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2016 Christian Hill
""" converting a spectrum to a colour

code from the web blog:
    https://scipython.com/blog/converting-a-spectrum-to-a-colour/

@author: Christian Hill
"""

import numpy as np
from scipy.constants import h, c, k
from rayoptics.util.line_intersection import intersect_with_3lines
from pathlib import Path


def xyz_from_xy(x, y):
    """Return the vector (x, y, 1-x-y)."""
    return np.array((x, y, 1-x-y))


class ColourSystem:
    """A class representing a colour system.

    A colour system defined by the CIE x, y and z=1-x-y coordinates of
    its three primary illuminants and its "white point".

    TODO: Implement gamma correction

    """

    # The CIE colour matching function for 380 - 780 nm in 5 nm intervals
    path = Path(__file__).resolve().parent
    cmf = np.loadtxt(path / 'cie-cmf.txt', usecols=(1, 2, 3))

    def __init__(self, red, green, blue, white):
        """Initialise the ColourSystem object.

        Pass vectors (ie NumPy arrays of shape (3,)) for each of the
        red, green, blue  chromaticities and the white illuminant
        defining the colour system.

        """

        # Chromaticities
        self.red, self.green, self.blue = red, green, blue
        self.white = white
        # The chromaticity matrix (rgb -> xyz) and its inverse
        self.M = np.vstack((self.red, self.green, self.blue)).T
        self.MI = np.linalg.inv(self.M)
        # White scaling array
        self.wscale = self.MI.dot(self.white)
        # xyz -> rgb transformation matrix
        self.T = self.MI / self.wscale[:, np.newaxis]

    def xyz_to_rgb(self, xyz, out_fmt=None):
        """Transform from xyz to rgb representation of colour.

        The output rgb components are normalized on their maximum
        value. If xyz is out the rgb gamut, it is desaturated until it
        comes into gamut.

        By default, fractional rgb components are returned; if
        out_fmt='html', the HTML hex string '#rrggbb' is returned.

        """

        rgb = self.T.dot(xyz)
        if np.any(rgb < 0):
            # We're not in the RGB gamut: apply relative colorimetric mapping
            rgb1 = self.relative_colorimetric_gamut_mapping(xyz)
            if np.any(rgb1 < -1.0e-14):
                # We're still not in the RGB gamut: approximate by desaturating
                print("still not in gamut: xyz, rgb0, rgb1", xyz, rgb, rgb1)
                w = - np.min(rgb1)
                rgb1 += w
            rgb = rgb1
        if not np.all(rgb == 0):
            # Normalize the rgb vector
            rgb /= np.max(rgb)

        if out_fmt == 'html':
            return self.rgb_to_hex(rgb)
        return rgb

    def rgb_to_hex(self, rgb):
        """Convert from fractional rgb values to HTML-style hex string."""

        hex_rgb = (255 * rgb).astype(int)
        return '#{:02x}{:02x}{:02x}'.format(*hex_rgb)

    def spec_to_xyz(self, spec):
        """Convert a spectrum to an xyz point.

        The spectrum must be on the same grid of points as the colour-matching
        function, self.cmf: 380-780 nm in 5 nm steps.

        """

        XYZ = np.sum(spec[:, np.newaxis] * self.cmf, axis=0)
        den = np.sum(XYZ)
        if den == 0.:
            return XYZ
        return XYZ / den

    def wvl_to_xyz(self, wv):
        """Convert a wavelength (nm) to an xyz point.

        The wavelength must be on the same grid of points as the
        colour-matching function, self.cmf: 380-780 nm in 5 nm steps.

        """
        index = int((wv-380.0)/5.0)
        if index < 0:
            return np.array(0., 0., 0.)

        try:
            XYZ = self.cmf[index]
        except IndexError:
            XYZ = np.array(0., 0., 0.)
        finally:
            den = np.sum(XYZ)
            if den == 0.:
                return XYZ
            return XYZ / den

    def spec_to_rgb(self, spec, out_fmt=None):
        """Convert a spectrum to an rgb value."""

        xyz = self.spec_to_xyz(spec)
        return self.xyz_to_rgb(xyz, out_fmt)

    def wvl_to_rgb(self, wv, out_fmt=None):
        """Convert a wavelength to an rgb value."""

        xyz = self.wvl_to_xyz(wv)
        return self.xyz_to_rgb(xyz, out_fmt)

    def relative_colorimetric_gamut_mapping(self, xyz):
        """ Apply relative colorimetric gamut mapping to XYZ and return RGB """
        xy_mapped = intersect_with_3lines(xyz[:2], self.white[:2],
                                          (self.blue[:2], self.green[:2]),
                                          (self.green[:2], self.red[:2]),
                                          (self.red[:2], self.blue[:2]))
        rgb = self.T.dot(xyz_from_xy(*xy_mapped))
        return rgb


illuminant_D65 = xyz_from_xy(0.3127, 0.3291)
cs_hdtv = ColourSystem(red=xyz_from_xy(0.67, 0.33),
                       green=xyz_from_xy(0.21, 0.71),
                       blue=xyz_from_xy(0.15, 0.06),
                       white=illuminant_D65)

cs_smpte = ColourSystem(red=xyz_from_xy(0.63, 0.34),
                        green=xyz_from_xy(0.31, 0.595),
                        blue=xyz_from_xy(0.155, 0.070),
                        white=illuminant_D65)

cs_srgb = ColourSystem(red=xyz_from_xy(0.64, 0.33),
                       green=xyz_from_xy(0.30, 0.60),
                       blue=xyz_from_xy(0.15, 0.06),
                       white=illuminant_D65)


def planck(lam, T):
    """ Returns the spectral radiance of a black body at temperature T.

    Returns the spectral radiance, B(lam, T), in W.sr-1.m-2 of a black body
    at temperature T (in K) at a wavelength lam (in nm), using Planck's law.

    """

    lam_m = lam / 1.e9
    fac = h*c/lam_m/k/T
    B = 2*h*c**2/lam_m**5 / (np.exp(fac) - 1)
    return B



""" package supplying utility functions for math and numpy support

    The ``rayoptics.util`` subpackage provides miscellaneous functions for
    geometric calculations, color calculations and anything else that doesn't
    have an obvious home.
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" tests for ideal imager

.. Created on Wed Jun 12 17:57:06 2019

.. codeauthor: Michael J. Hayford
"""

import math
import pytest
from rayoptics.optical.idealimager import ideal_imager_setup


def test_ideal_imager_setup(**kwargs):
    m1 = -0.5
    s1 = -10.
    m1s1 = ideal_imager_setup(m=m1, s=s1)
    sp1 = m1s1.sp
    tt1 = m1s1.tt
    f1 = m1s1.f
    assert ideal_imager_setup(m=m1, s=s1) == ideal_imager_setup(s=s1, m=m1)
    assert ideal_imager_setup(m=m1, sp=sp1) == ideal_imager_setup(sp=sp1, m=m1)
    assert ideal_imager_setup(m=m1, tt=tt1) == ideal_imager_setup(tt=tt1, m=m1)
    assert ideal_imager_setup(m=m1, f=f1) == ideal_imager_setup(f=f1, m=m1)

    tt2 = 4.0
    f2 = 1.0
    tt2f2 = ideal_imager_setup(tt=tt2, f=f2)
    assert tt2f2.m == -1
    assert tt2f2.s == -2
    assert tt2f2.sp == 2
    f3 = 2.0
    with pytest.raises(ValueError):
        ideal_imager_setup(tt=tt2, f=f3)

    s_inf = -math.inf
    efl = 25.0
    s_inf_efl = ideal_imager_setup(s=s_inf, f=efl)
    assert s_inf_efl.m == -0
    assert s_inf_efl.s == -math.inf
    assert s_inf_efl.sp == efl



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Oct 25 12:16:36 2018

@author: Mike
"""
import math
import numpy as np
from pathlib import Path
import csv
import timeit

# ray-optics
import rayoptics as ro
from rayoptics.optical.opticalmodel import open_model
import rayoptics.optical.raytrace as rt
from rayoptics.util.misc_math import normalize


def setup(filename):
    root_pth = Path(ro.__file__).resolve().parent
    opm = open_model(root_pth/filename)
    sm = opm.seq_model
    osp = opm.optical_spec
    fld, wvl, foc = osp.lookup_fld_wvl_focus(1)
    vig_pupil = fld.apply_vignetting([0.5, 0.5])
    fod = osp.parax_data.fod
    eprad = fod.enp_radius
    pt1 = np.array([eprad*vig_pupil[0], eprad*vig_pupil[1],
                    fod.obj_dist+fod.enp_dist])
    pt0 = osp.obj_coords(fld)
    dir0 = normalize(pt1 - pt0)
    return (sm, pt0, dir0, wvl)


def run_test(tst_name, setup_stmt, trials, repeat=5, file=None):
    output = '{:s} - {} trials: min time {:.2f}, ' \
             'max time {:.2f}, spread {:.2f}%'
    t = timeit.repeat('ray=rt.trace(*trace_args)', setup=setup_stmt,
                      number=trials, repeat=repeat, globals=globals())
    pcnt_sprd = 100*(max(t) - min(t))/min(t)
    print(output.format(tst_name, trials, min(t), max(t), pcnt_sprd),
          file=file)
    return [tst_name, trials, min(t), max(t), pcnt_sprd, t]


if __name__ == '__main__':
    trials10 = 10000
    trials15 = 15000
    trials20 = 20000
    trials25 = 25000
    trials30 = 30000
    trials40 = 40000
    trials50 = 50000
    trials80 = 80000
    trials100 = 100000
    setup_str = 'trace_args=setup("{:s}")'

    results = []
    rpt = 5

    root_pth = Path(ro.__file__).resolve().parent
    with open(root_pth/'optical/tests/trace_results.txt', mode='w') as f:

        tst_name = 'singlet'
        trials = trials100
        setup_stmt = setup_str.format("codev/tests/singlet.seq")
        results.append(run_test(tst_name, setup_stmt, trials, file=f,
                                repeat=rpt))

        tst_name = 'landscape lens'
        trials = trials80
        setup_stmt = setup_str.format("codev/tests/landscape_lens.seq")
        results.append(run_test(tst_name, setup_stmt, trials, file=f,
                                repeat=rpt))

        tst_name = 'Sasian triplet'
        trials = trials50
        setup_stmt = setup_str.format("../models/Sasian Triplet.roa")
        results.append(run_test(tst_name, setup_stmt, trials, file=f,
                                repeat=rpt))

        tst_name = 'double gauss'
        trials = trials30
        setup_stmt = setup_str.format("codev/tests/ag_dblgauss.seq")
        results.append(run_test(tst_name, setup_stmt, trials, file=f,
                                repeat=rpt))

        tst_name = '2 spherical mirrors (spheres)'
        trials = trials100
        setup_stmt = setup_str.format("../models/TwoSphericalMirror.roa")
        results.append(run_test(tst_name, setup_stmt, trials, file=f,
                                repeat=rpt))

        tst_name = '2 spherical mirrors (conics)'
        trials = trials100
        setup_stmt = setup_str.format("../models/TwoMirror.roa")
        results.append(run_test(tst_name, setup_stmt, trials, file=f,
                                repeat=rpt))

        tst_name = 'paraboloid'
        trials = trials100
        setup_stmt = setup_str.format("codev/tests/paraboloid.seq")
        results.append(run_test(tst_name, setup_stmt, trials, file=f,
                                repeat=rpt))

        tst_name = 'Cassegrain'
        trials = trials100
        setup_stmt = setup_str.format("../models/Cassegrain.roa")
        results.append(run_test(tst_name, setup_stmt, trials, file=f,
                                repeat=rpt))

        tst_name = 'Ritchey-Chretien'
        trials = trials100
        setup_stmt = setup_str.format("../models/Ritchey_Chretien.roa")
        results.append(run_test(tst_name, setup_stmt, trials, file=f,
                                repeat=rpt))

        tst_name = 'cell phone camera'
        trials = trials10
        setup_stmt = setup_str.format("optical/tests/cell_phone_camera.roa")
        results.append(run_test(tst_name, setup_stmt, trials, file=f,
                                repeat=rpt))

    with open(root_pth/'optical/tests/trace_data.csv', mode='w') as f:
        w = csv.writer(f, delimiter=',', quotechar='"',
                       quoting=csv.QUOTE_MINIMAL)
        for tst in results:
            w.writerow(tst[:5])

    with open(root_pth/'optical/tests/trace_data.txt', mode='w') as f:
        print(results, file=f)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Top level model classes

.. Created on Wed Mar 14 11:08:28 2018

.. codeauthor: Michael J. Hayford
"""
import os.path
import json_tricks

import rayoptics

from rayoptics.optical import elements
import rayoptics.optical.model_constants as mc

from rayoptics.optical.elements import ElementModel
from rayoptics.optical.paraxialdesign import ParaxialModel
from rayoptics.optical.sequential import SequentialModel
from rayoptics.optical.opticalspec import OpticalSpecs
from rayoptics.optical.specsheet import create_specsheet_from_model
from rayoptics.optical.model_enums import DimensionType as dt


class SystemSpec:
    """ Container for units and other system level constants

    Attributes:
        title (str): a short description of the model
        initials (str): user initials or other id
        dimensions: a DimensionType enum of the model linear units
        temperature (float): model temperature in degrees Celsius
        pressure (float): model pressure in mm/Hg
    """
    def __init__(self):
        self.title = ''
        self.initials = ''
        self.dimensions = dt.MM
        self.temperature = 20.0
        self.pressure = 760.0

    def nm_to_sys_units(self, nm):
        """ convert nm to system units

        Args:
            nm (float): value in nm

        Returns:
            float: value converted to system units
        """
        if self.dimensions == dt.M:
            return 1e-9 * nm
        elif self.dimensions == dt.CM:
            return 1e-7 * nm
        elif self.dimensions == dt.MM:
            return 1e-6 * nm
        elif self.dimensions == dt.IN:
            return 1e-6 * nm/25.4
        elif self.dimensions == dt.FT:
            return 1e-6 * nm/304.8
        else:
            return nm


class OpticalModel:
    """ Top level container for optical model.

    The OpticalModel serves as a top level container of model properties.
    Key aspects are built-in element and surface based repesentations of the
    optical surfaces.
    A sequential optical model is a sequence of surfaces and gaps.
    Additionally, it includes optical usage information to specify the
    aperture, field of view, spectrum and focus.
    """

    def __init__(self, radius_mode=False, specsheet=None):
        self.ro_version = rayoptics.__version__
        self.radius_mode = radius_mode
        self.specsheet = specsheet
        self.system_spec = SystemSpec()
        self.seq_model = SequentialModel(self)
        self.optical_spec = OpticalSpecs(self)
        self.parax_model = ParaxialModel(self)
        self.ele_model = ElementModel(self)

        if self.specsheet:
            self.set_from_specsheet()

    def name(self):
        return self.system_spec.title

    def reset(self):
        rdm = self.radius_mode
        self.__init__()
        self.radius_mode = rdm

    def __json_encode__(self):
        attrs = dict(vars(self))
        if hasattr(self, 'app_manager'):
            del attrs['app_manager']
        return attrs

    def set_from_specsheet(self, specsheet=None):
        if specsheet:
            self.specsheet = specsheet
        else:
            specsheet = self.specsheet
        self.seq_model.set_from_specsheet(specsheet)
        self.optical_spec.set_from_specsheet(specsheet)

    def save_model(self, file_name):
        file_extension = os.path.splitext(file_name)[1]
        filename = file_name if len(file_extension) > 0 else file_name+'.roa'
        fs_dict = {}
        fs_dict['optical_model'] = self
        with open(filename, 'w') as f:
            json_tricks.dump(fs_dict, f, indent=1,
                             separators=(',', ':'), allow_nan=True)

    def sync_to_restore(self):
        if not hasattr(self, 'ro_version'):
            self.ro_version = rayoptics.__version__

        self.seq_model.sync_to_restore(self)
        self.ele_model.sync_to_restore(self)
        self.optical_spec.sync_to_restore(self)

        if hasattr(self, 'parax_model'):
            self.parax_model.sync_to_restore(self)
        else:
            self.parax_model = ParaxialModel(self)

        if hasattr(self, 'specsheet'):
            self.specsheet.sync_to_restore(self)
        else:
            self.specsheet = None

        self.update_model()

    def update_model(self):
        self.seq_model.update_model()
        self.optical_spec.update_model()
        self.parax_model.update_model()
        self.ele_model.update_model()
        if self.specsheet is None:
            self.specsheet = create_specsheet_from_model(self)

    def nm_to_sys_units(self, nm):
        """ convert nm to system units

        Args:
            nm (float): value in nm

        Returns:
            float: value converted to system units
        """
        return self.system_spec.nm_to_sys_units(nm)

    def add_lens(self, **kwargs):
        seq, ele = elements.create_lens(**kwargs)
        self.insert_ifc_gp_ele(seq, ele, **kwargs)

    def add_mirror(self, **kwargs):
        seq, ele = elements.create_mirror(**kwargs)
        self.insert_ifc_gp_ele(seq, ele, **kwargs)

    def add_thinlens(self, **kwargs):
        seq, ele = elements.create_thinlens(**kwargs)
        self.insert_ifc_gp_ele(seq, ele, **kwargs)

    def add_dummy_plane(self, **kwargs):
        seq, ele = elements.create_dummy_plane(**kwargs)
        self.insert_ifc_gp_ele(seq, ele, **kwargs)

    def insert_ifc_gp_ele(self, *args, **kwargs):
        """ insert interfaces and gaps into seq_model and eles into ele_model
        """
        seq, ele = args
        if 'idx' in kwargs:
            self.seq_model.cur_surface = kwargs['idx']

        # distinguish between adding a new chunk, which requires splitting a
        #  gap in two, and replacing a node, which uses the existing gaps.
        if 'insert' in kwargs:
            t = kwargs['t'] if 't' in kwargs else 0.
            g, ag = elements.create_air_gap(t=t, ref_ifc=seq[-1][mc.Intfc])
            seq[-1][mc.Gap] = g
            ele.append(ag)
        else:
            # replacing an existing node. need to hook new chunk final
            # interface to the existing gap and following (air gap) element
            g = self.seq_model.gaps[self.seq_model.cur_surface+1]
            seq[-1][mc.Gap] = g
            ag = self.ele_model.gap_dict[g]
            ag.ref_ifc = seq[-1][mc.Intfc]  # tacit assumption is ag == AirGap

        for sg in seq:
            self.seq_model.insert(sg[mc.Intfc], sg[mc.Gap])

        for e in ele:
            self.ele_model.add_element(e)
        self.ele_model.sequence_elements()

    def remove_ifc_gp_ele(self, *args, **kwargs):
        """ remove interfaces and gaps from seq_model and eles from ele_model
        """
        seq, ele = args
        sg = seq[0]
        idx = self.seq_model.ifcs.index(sg[mc.Intfc])

        # verify that the sequences match
        seq_match = True
        for i, sg in enumerate(seq):
            if sg[0] is not self.seq_model.ifcs[idx+i]:
                seq_match = False
                break

        if seq_match:
            # remove interfaces in reverse
            for i in range(idx+len(seq)-1, idx-1, -1):
                self.seq_model.remove(i)

        for e in ele:
            self.ele_model.remove_element(e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Functions to read a CODE V .seq file and populate a sequential model

.. Created on Tue Jan 16 10:14:12 2018

.. codeauthor: Michael J. Hayford
"""
import logging
import math

from . import tla
from . import reader as cvr

from rayoptics.optical.opticalmodel import OpticalModel
from rayoptics.optical import model_enums
from rayoptics.optical.model_enums import PupilType, FieldType
from rayoptics.optical.model_enums import DimensionType as dt
from rayoptics.optical.model_enums import DecenterType as dec
from rayoptics.optical.surface import (DecenterData, Circular, Rectangular,
                                       Elliptical)
from rayoptics.optical import profiles
from rayoptics.optical import doe
from rayoptics.optical.medium import Air, Glass, InterpolatedGlass
from rayoptics.optical.opticalspec import Field
from rayoptics.util.misc_math import isanumber

from opticalglass import glassfactory as gfact
from opticalglass import glasserror as ge

_tla = tla.MapTLA()

# Support for CODE V private catalog materials. 
# Should this be in OpticalModel?
_reading_private_catalog = False
_private_catalog_wvls = None
_private_catalog_glasses = {}


def fictitious_glass_decode(gc):
    ''' glass code parser, allowing for greater precision of n and v '''
    rindex_part = int(gc)
    magnitude = int(math.floor(math.log10(rindex_part))) + 1
    n = 1.0 + rindex_part/10**magnitude
    v = round(100.0*(gc - int(gc)), 6)
    return n, v


def read_lens(filename):
    ''' given a CODE V .seq filename, return an OpticalModel  '''
    global _reading_private_catalog
    logging.basicConfig(filename='cv_cmd_proc.log',
                        filemode='w',
                        level=logging.DEBUG)
    _reading_private_catalog = False
    opt_model = OpticalModel()
    cmds = cvr.read_seq_file(filename)
    for i, c in enumerate(cmds):
        cmd_fct, tla, qlist, dlist = process_command(c)
        if cmd_fct:
            eval_str = cmd_fct + '(opt_model, tla, qlist, dlist)'
            eval(eval_str)
        else:
            logging.info('Line %d: Command %s not supported', i+1, c[0])

    opt_model.update_model()

    return opt_model


def process_command(cmd):
    global _reading_private_catalog
    CmdFct, IndxQuals, DataType, Quals = range(4)
    tla = cmd[0][:3].upper()
    qlist = []
    dlist = []
    cmd_fct = None
    cmd_def = _tla.find(tla)
    if cmd_def:
        cmd_fct = cmd_def[CmdFct]
        iquals = cmd_def[IndxQuals]
        quals = cmd_def[Quals]
        data_type = cmd_def[DataType]
        data_found = False
        for t in cmd[1:]:
            if not data_found:
                if t in quals:
                    qlist.append((t,))
                elif t[:1] in iquals:
                    qlist.append((t[:1], t[1:]))
                else:
                    data_found = True
            if data_found:
                if data_type == 'String':
                    dlist.append(t)
                elif data_type == 'Double':
                    dlist.append(eval("float("+t+")"))
                elif data_type == 'Integer':
                    dlist.append(eval("int("+t+")"))
                elif data_type == 'Boolean':
                    if t[:1].upper() == 'N':
                        dlist.append(False)
                    else:
                        dlist.append(True)
    elif tla[:1] == 'S':
        cmd_fct = 'surface_cmd'
        qlist.append((tla[:1], tla[1:]))
        tla = 'S'
        dlist.append(float(cmd[1]))  # radius/curvature
        dlist.append(float(cmd[2]))  # thickness
        cmd_len = len(cmd)
        if cmd_len > 3:
            dlist.append(cmd[3])  # glass
        if cmd_len > 4:
            dlist.append(cmd[4])  # rmd
    elif _reading_private_catalog and isinstance(cmd[0], str):
        global _private_catalog_wvls, _private_catalog_glasses
        label = cmd[0]
        for t in cmd[1:]:
            dlist.append(eval("float("+t+")"))
        prv_glass = InterpolatedGlass(label, wvls=_private_catalog_wvls,
                                      rndx=dlist)
        _private_catalog_glasses[label] = prv_glass

    return cmd_fct, tla, qlist, dlist


def log_cmd(label, tla, qlist, dlist):
    logging.debug("%s: %s %s %s", label, tla, str(qlist), str(dlist))


def wvl_spec_data(optm, tla, qlist, dlist):
    osp = optm.optical_spec
    if tla == "WL":
        osp.spectral_region.wavelengths = dlist
        osp.spectral_region.calc_colors()
    elif tla == "WTW":
        osp.spectral_region.spectral_wts = dlist
    elif tla == "REF":
        osp.spectral_region.reference_wvl = dlist[0]-1
    elif tla == "CWL":
        osp.spectral_region.coating_wvl = dlist


def pupil_spec_data(optm, tla, qlist, dlist):
    osp = optm.optical_spec
    osp.pupil.key = model_enums.get_ape_key_for_type(PupilType[tla])
    osp.pupil.value = dlist[0]
    logging.debug("pupil_spec_data: %s %f", tla, dlist[0])


def field_spec_data(optm, tla, qlist, dlist):
    fov = optm.optical_spec.field_of_view
    if tla == 'XOB' or tla == 'YOB':
        field_type = FieldType.OBJ_HT
    elif tla == 'XAN' or tla == 'YAN':
        field_type = FieldType.OBJ_ANG
    elif tla == 'XIM' or tla == 'YIM':
        field_type = FieldType.IMG_HT
    else:
        field_type = None

    if field_type:
        fov.key = model_enums.get_fld_key_for_type(field_type)

    if len(fov.fields) != len(dlist):
        fov.fields = [Field() for f in range(len(dlist))]

    if tla[0] == 'V':
        attr = tla.lower()
    elif tla[0] == 'X' or tla[0] == 'Y':
        attr = tla[0].lower()
    elif tla == 'WTF':
        attr = 'wt'

    for i, f in enumerate(fov.fields):
        f.__setattr__(attr, dlist[i])

    log_cmd("field_spec_data", tla, qlist, dlist)


def spec_data(optm, tla, qlist, dlist):
    if tla == "LEN":
        optm.reset()
    elif tla == "RDM":
        if len(dlist) == 0:
            optm.radius_mode = True
        else:
            optm.radius_mode = dlist[0]
    elif tla == "TIT":
        optm.system_spec.title = dlist[0]
    elif tla == "INI":
        optm.system_spec.initials = dlist[0]
    elif tla == "DIM":
        dim = dlist[0].upper()
        if dim == 'M':
            dim = dt.MM
        elif dim == 'C':
            dim = dt.CM
        elif dim == 'I':
            dim = dt.IN
        optm.system_spec.dimensions = dim
    elif tla == "TEM":
        optm.system_spec.temperature = dlist[0]
    elif tla == "PRE":
        optm.system_spec.pressure = dlist[0]
    log_cmd("spec_data", tla, qlist, dlist)


def get_index_qualifier(seq_model, qtype, qlist):
    def num_or_alpha(idx):
        if idx.isdigit():
            return int(idx)
        elif idx.isalpha():
            if idx == 'O':
                return 0
            elif idx == 'I':
                return seq_model.get_num_surfaces()-1
            elif idx == 'S':
                return seq_model.stop_surface
        elif idx.isascii():
            if qtype == 'L':
                return idx
        else:
            return None
    for q in qlist:
        if q[0] == qtype:
            if q[1].find('..') > 0:
                i = q[1].find('..')
                return (num_or_alpha(q[1][:i]), num_or_alpha(q[1][i+2:]))
            else:
                return num_or_alpha(q[1]),


def surface_cmd(opt_model, tla, qlist, dlist):
    seq_model = opt_model.seq_model
    idx, = get_index_qualifier(seq_model, 'S', qlist)
    update_surface_and_gap(opt_model, dlist, idx)


def update_surface_and_gap(opt_model, dlist, idx=None):
    seq_model = opt_model.seq_model
    if isinstance(idx, int):
        s, g = seq_model.get_surface_and_gap(idx)
        seq_model.set_cur_surface(idx)
    else:
        s, g = seq_model.insert_surface_and_gap()

    if opt_model.radius_mode:
        if dlist[0] != 0.0:
            s.profile.cv = 1.0/dlist[0]
        else:
            s.profile.cv = 0.0
    else:
        s.profile.cv = dlist[0]

    if g:
        g.thi = dlist[1]

        if len(dlist) < 3:
            g.medium = Air()
        else:
            if dlist[2].upper() == 'REFL':
                s.interact_mode = 'reflect'
                g.medium = seq_model.gaps[seq_model.cur_surface-1].medium
            else:
                g.medium = process_glass_data(dlist[2])

    else:
        # at image surface, apply defocus to previous thickness
        seq_model.gaps[idx-1].thi += dlist[1]


def process_glass_data(glass_data):
    """ process GLA string for fictitious, catalog or private catalog glass """
    if isanumber(glass_data):  # process as fictitious glass code
        n, v = fictitious_glass_decode(float(glass_data))
        medium = Glass(n, v, '')

    else:  # look for glass name and optional catalog
        name_cat = glass_data.split('_')
        if len(name_cat) == 2:
            name, cat = name_cat
        elif len(name_cat) == 1:
            name, cat = glass_data, None
        if cat is not None:
            if cat.upper() == 'SCHOTT' and name[:1].upper() == 'N':
                name = name[:1]+'-'+name[1:]
            elif cat.upper() == 'OHARA' and name[:1].upper() == 'S':
                name = name[:1]+'-'+name[1:]
                if not name[-2:].isdigit() and name[-1].isdigit():
                    name = name[:-1]+' '+name[-1]
            try:
                medium = gfact.create_glass(name, cat)
            except ge.GlassNotFoundError as gerr:
                logging.info('%s glass data type %s not found',
                             gerr.catalog,
                             gerr.name)
                logging.info('Replacing material with air.')
                medium = Air()
        else:  # name with no data. default to crown glass
            global _private_catalog_glasses
            if name in _private_catalog_glasses:
                medium = _private_catalog_glasses[name]
            else:
                medium = Glass(1.5, 60.0, name)
    return medium


def private_catalog(optm, tla, qlist, dlist):
    global _reading_private_catalog, _private_catalog_wvls
    if tla == "PRV":
        _reading_private_catalog = True
    elif tla == "PWL":
        _private_catalog_wvls = dlist
    elif tla == "END":
        _reading_private_catalog = False
        _private_catalog_wvls = None

    log_cmd("private_catalog", tla, qlist, dlist)


def surface_data(optm, tla, qlist, dlist):
    seq_model = optm.seq_model
    idx = get_index_qualifier(seq_model, 'S', qlist)
    if not idx:
        idx = seq_model.cur_surface
    if tla == 'SLB':
        seq_model.ifcs[idx].label = dlist[0]
    elif tla == 'STO':
        seq_model.stop_surface = idx
    elif tla == 'THI':
        seq_model.gaps[idx].thi = dlist[0]
    elif tla == 'SPH':
        update_surface_profile(seq_model, 'Spherical', idx)
    elif tla == 'CON':
        update_surface_profile(seq_model, 'Conic', idx)
    elif tla == 'ASP':
        update_surface_profile(seq_model, 'EvenPolynomial', idx)
    log_cmd("surface_data", tla, qlist, dlist)


def update_surface_profile(seq_model, profile_type, idx=None):
    if not isinstance(idx, int):
        idx = seq_model.cur_surface
    cur_profile = seq_model.ifcs[idx].profile
    new_profile = profiles.mutate_profile(cur_profile, profile_type)
    seq_model.ifcs[idx].profile = new_profile
    return seq_model.ifcs[idx].profile


def profile_data(optm, tla, qlist, dlist):
    seq_model = optm.seq_model
    idx = get_index_qualifier(seq_model, 'S', qlist)
    if not idx:
        idx = seq_model.cur_surface
    if tla == 'CUX':
        seq_model.ifcs[idx].profile.cv = dlist[0]
    elif tla == 'CUY':
        seq_model.ifcs[idx].profile.cv = dlist[0]
    elif tla == 'RDX':
        if dlist[0] != 0.0:
            seq_model.ifcs[idx].profile.cv = 1.0/dlist[0]
        else:
            seq_model.ifcs[idx].profile.cv = 0.0
    elif tla == 'RDY':
        if dlist[0] != 0.0:
            seq_model.ifcs[idx].profile.cv = 1.0/dlist[0]
        else:
            seq_model.ifcs[idx].profile.cv = 0.0
    elif tla == 'K':
        seq_model.ifcs[idx].profile.cc = dlist[0]
    elif tla == 'A':
        seq_model.ifcs[idx].profile.coef4 = dlist[0]
    elif tla == 'B':
        seq_model.ifcs[idx].profile.coef6 = dlist[0]
    elif tla == 'C':
        seq_model.ifcs[idx].profile.coef8 = dlist[0]
    elif tla == 'D':
        seq_model.ifcs[idx].profile.coef10 = dlist[0]
    elif tla == 'E':
        seq_model.ifcs[idx].profile.coef12 = dlist[0]
    elif tla == 'F':
        seq_model.ifcs[idx].profile.coef14 = dlist[0]
    elif tla == 'G':
        seq_model.ifcs[idx].profile.coef16 = dlist[0]
    elif tla == 'H':
        seq_model.ifcs[idx].profile.coef18 = dlist[0]
    elif tla == 'J':
        seq_model.ifcs[idx].profile.coef20 = dlist[0]
    log_cmd("profile_data", tla, qlist, dlist)


def aperture_data(opm, tla, qlist, dlist):
    """ add aperture data, either creating a new aperture or modifying the last """
    seq_model = opm.seq_model
    idx = get_index_qualifier(seq_model, 'S', qlist)
    lbl = get_index_qualifier(seq_model, 'L', qlist)
    if not idx:
        idx = seq_model.cur_surface
    ca_type = tla[0]
    data_type = tla[2]
    ifc = seq_model.ifcs[idx]
    ca_list = ifc.clear_apertures

    for q in qlist:
        if q[0] == 'EDG':
            ca_list = ifc.edge_apertures
        elif q[0] == 'HOL':
            log_cmd("aperture_data", tla, qlist, dlist)
#            ca_list = ifc.holes
            return
        elif q[0] == 'OBS':
            log_cmd("aperture_data", tla, qlist, dlist)
            return

    if len(ca_list) == 0 or ca_type != type(ca_list[-1]).__name__[0]:
        if ca_type == 'C':
            ca = Circular()
        elif ca_type == 'R':
            ca = Rectangular()
        elif ca_type == 'E':
            ca = Elliptical()
        ca_list.append(ca)
    else:
        ca = ca_list[-1]

    if data_type == 'R':
        ca.radius = dlist[0]
    elif data_type == 'X':
        ca.x_half_width = dlist[0]
    elif data_type == 'Y':
        ca.y_half_width = dlist[0]

    log_cmd("aperture_data", tla, qlist, dlist)


def aperture_data_general(opm, tla, qlist, dlist):
    """ handle the general aperture commands, add to end of list """
    seq_model = opm.seq_model
    idx = get_index_qualifier(seq_model, 'S', qlist)
    if not idx:
        idx = seq_model.cur_surface
    ca_type = tla[0]

    if ca_type == 'C':
        ca = Circular(radius=dlist[0], x_offset=dlist[1], y_offset=dlist[2],
                      rotation=dlist[3])
    elif ca_type == 'R':
        ca = Rectangular(x_half_width=dlist[0], y_half_width=dlist[1],
                         x_offset=dlist[2], y_offset=dlist[3],
                         rotation=dlist[3])
    elif ca_type == 'E':
        ca = Elliptical(x_half_width=dlist[0], y_half_width=dlist[1],
                        x_offset=dlist[2], y_offset=dlist[3],
                        rotation=dlist[3])

    seq_model.ifcs[idx].clear_apertures.append(ca)

    log_cmd("aperture_data_general", tla, qlist, dlist)


def aperture_offset(opm, tla, qlist, dlist):
    """ handle the aperture offset commands, assume last aperture in list """
    seq_model = opm.seq_model
    idx = get_index_qualifier(seq_model, 'S', qlist)
    if not idx:
        idx = seq_model.cur_surface
    ca = seq_model.ifcs[idx].clear_apertures[-1]
    offset_type = tla[2]

    if offset_type == 'X':
        ca.x_offset = dlist[0]
    elif offset_type == 'Y':
        ca.y_offset = dlist[0]
    elif offset_type == 'O':
        ca.rotation = dlist[0]

    log_cmd("aperture_offset", tla, qlist, dlist)


def decenter_data(optm, tla, qlist, dlist):
    seq_model = optm.seq_model
    idx = get_index_qualifier(seq_model, 'S', qlist)
    if not idx:
        idx = seq_model.cur_surface
    ifc = seq_model.ifcs[idx]

    if not ifc.decenter:
        ifc.decenter = DecenterData(dec.LOCAL)

    decenter = ifc.decenter
    if tla == 'XDE':
        decenter.dec[0] = dlist[0]
    elif tla == 'YDE':
        decenter.dec[1] = dlist[0]
    elif tla == 'ZDE':
        decenter.dec[2] = dlist[0]
    elif tla == 'ADE':
        decenter.euler[0] = dlist[0]
    elif tla == 'BDE':
        decenter.euler[1] = dlist[0]
    elif tla == 'CDE':
        decenter.euler[2] = dlist[0]
    elif tla == 'DAR':
        decenter.dtype = dec.DAR
    elif tla == 'BEN':
        decenter.dtype = dec.BEND
    elif tla == 'REV':
        decenter.dtype = dec.REV

    decenter.update()

    log_cmd("decenter_data", tla, qlist, dlist)


#  DIF DOE
#  HOR 1.0
#  HWL 587.5618; HCT R
#  HCO C1 -0.001807322521767816; HCC C1 100
def diffractive_optic(optm, tla, qlist, dlist):
    seq_model = optm.seq_model
    idx = get_index_qualifier(seq_model, 'S', qlist)
    if not idx:
        idx = seq_model.cur_surface
    ifc = seq_model.ifcs[idx]

    if tla == "DIF":
        for q in qlist:
            if "DOE" == q[0]:
                ifc.phase_element = doe.DiffractiveElement()
    elif tla == "HOR":
        if hasattr(ifc, 'phase_element'):
            ifc.phase_element.order = dlist[0]
    elif tla == "HWL":
        if hasattr(ifc, 'phase_element'):
            ifc.phase_element.ref_wl = dlist[0]
    elif tla == "HCT":
        if hasattr(ifc, 'phase_element'):
            for q in qlist:
                if "R" == q[0]:
                    ifc.phase_element.phase_fct = doe.radial_phase_fct
    elif tla == "HCO":
        cidx = get_index_qualifier(seq_model, 'C', qlist)[0]
        if hasattr(ifc, 'phase_element'):
            num_coefs = len(ifc.phase_element.coefficients)
            coefs = ifc.phase_element.coefficients
            if cidx <= num_coefs:
                coefs[cidx-1] = dlist[0]
            elif cidx == num_coefs + 1:
                coefs.append(dlist[0])
            else:
                coefs.extend([0.]*(cidx - num_coefs))
                coefs[cidx-1] = dlist[0]

    log_cmd("diffractive_optic", tla, qlist, dlist)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
"""

.. Created on Thu Oct 10 21:57:25 2019

.. codeauthor: Michael J. Hayford
"""

import logging
from collections import namedtuple

import numpy as np
from matplotlib import lines
from matplotlib import patches
from matplotlib import widgets

from rayoptics.mpl.styledfigure import StyledFigure

from rayoptics.gui import util
from rayoptics.util import rgb2mpl


SelectInfo = namedtuple('SelectInfo', ['artist', 'info'])
""" tuple grouping together an artist and info returned from contains(event)

    Attributes:
        artist: the artist
        info: a dictionary of artist specific details of selection
"""


def display_artist_and_event(callback_str, event, artist):
    if artist and hasattr(artist, 'shape'):
        shape, handle = artist.shape
        print('{} {}: shape: {}.{}'.format(callback_str, event.name,
                                           shape.get_label(), handle))
    else:
        print('{} {}: shape: None'.format(callback_str, event.name))


class InteractiveFigure(StyledFigure):
    """ Editable version of optical system layout, aka Live Layout

    Attributes:
        opt_model: parent optical model
        do_draw_frame: if True, draw frame around the figure
        do_draw_axes: if True, draw coordinate axes for the figure
        oversize_factor: what fraction to oversize the view bounding box
        aspect: 'equal' for 1:1 aspect ratio, 'auto' for best ratio
    """

    def __init__(self,
                 do_draw_frame=False,
                 do_draw_axes=False,
                 oversize_factor=0.05,
                 aspect='equal',
                 view_bbox=None,
                 do_scale_bounds=False,
                 **kwargs):
        self.linewidth = 0.5
        self.do_draw_frame = do_draw_frame
        self.do_draw_axes = do_draw_axes
        self.oversize_factor = oversize_factor
        self.aspect = aspect
        self.hilited = None
        self.selected = None
        self.do_scale_bounds = do_scale_bounds

        self.do_action = self.do_shape_action
        self.event_dict = {}

        self.on_finished = None

        super().__init__(**kwargs)

        self.update_data()
        self.view_bbox = view_bbox if view_bbox else self.fit_axis_limits()

    def connect_events(self, action_dict=None):
        'connect to all the events we need'
        if action_dict is None:
            action_dict = {'motion_notify_event': self.on_motion,
                           'button_press_event': self.on_press,
                           'button_release_event': self.on_release,
                           # 'button_press_event': self.display_event,
                           # 'button_release_event': self.display_event,
                           # 'pick_event': self.on_select,
                           }
        self.callback_ids = []
        for event, action in action_dict.items():
            self.event_dict[event] = action
            cid = self.canvas.mpl_connect(event, action)
            self.callback_ids.append(cid)

    def disconnect_events(self):
        'disconnect all the stored connection ids'
        for clbk in self.callback_ids:
            self.canvas.mpl_disconnect(clbk)
        self.callback_ids = None
        event_dict, self.event_dict = self.event_dict, {}
        return event_dict

    @property
    def is_unit_aspect_ratio(self):
        return self.aspect == 'equal'

    @is_unit_aspect_ratio.setter
    def is_unit_aspect_ratio(self, value):
        self.aspect = 'equal' if value else 'auto'

    def refresh(self):
        self.update_data()
        self.plot()

    def update_data(self):
        pass

    def action_complete(self):
        if self.on_finished:
            self.on_finished()
            self.on_finished = None

    def register_action(self, *args, **kwargs):
        action_obj = args[0]
        fig = kwargs.pop('figure')

        def do_command_action(event, target, event_key):
            nonlocal action_obj, fig
            try:
                action_obj.actions[event_key](fig, event)
            except KeyError:
                pass
        self.do_action = do_command_action

    def register_pan(self, on_finished):
        action_obj = PanAction()
        self.register_action(action_obj, figure=self)
        self.on_finished = on_finished

    def register_zoom_box(self, on_finished):
        self.zoom_box_action = ZoomBoxAction(self)

        def do_command_action(event, target, event_key):
            pass

        self.do_action = do_command_action
        self.on_finished = on_finished

    def update_patches(self, shapes):
        """ loop over the input shapes, fetching their current geometry
        and attaching it to the corresponding ``Artist``
        """
        bbox_list = []
        for shape in shapes:
            handles = shape.update_shape(self)
            for key, gui_handle in handles.items():
                poly, bbox = gui_handle
                # add shape and handle key as attribute on artist
                poly.shape = (shape, key)
                self.artists.append(poly)
                if len(bbox_list) == 0:
                    bbox_list = bbox
                else:
                    bbox_list = np.vstack((bbox_list, bbox))
        bbox = util.bbox_from_poly(bbox_list)
        return bbox

    def create_polygon(self, poly, rgb_color, **kwargs):
        def highlight(p):
            fc = p.get_facecolor()
            ec = p.get_edgecolor()
            lw = p.get_linewidth()
            p.unhilite = (fc, ec, lw)
            alpha = fc[3]+0.5
            if alpha > 1.0:
                alpha -= 1.0  # subtract 0.5 instead of adding
            p.set_facecolor((fc[0], fc[1], fc[2], alpha))

        def unhighlight(p):
            fc, ec, lw = p.unhilite
            p.set_facecolor(fc)
            p.set_edgecolor(ec)
            p.set_linewidth(lw)
            p.unhilite = None

        if 'linewidth' not in kwargs:
            kwargs['linewidth'] = self.linewidth
        fill_color = kwargs.pop('fill_color', rgb_color)
        if not isinstance(fill_color, str):
            fill_color = rgb2mpl.rgb2mpl(fill_color)
        p = patches.Polygon(poly, closed=True, fc=fill_color,
                            ec='black', **kwargs)
        p.highlight = highlight
        p.unhighlight = unhighlight
        return p

    def create_polyline(self, poly, **kwargs):
        def highlight(p):
            lw = p.get_linewidth()
            ls = p.get_linestyle()
            c = p.get_color()
            p.unhilite = (c, lw, ls)
            # hilite_lw = 2
            hilite_lw = 5
            p.set_linewidth(hilite_lw)
            if hilite_color:
                p.set_color(hilite_color)

        def unhighlight(p):
            c, lw, ls = p.unhilite
            p.set_linewidth(lw)
            p.set_linestyle(ls)
            p.set_color(c)
            p.unhilite = None

        x = poly.T[0]
        y = poly.T[1]
        hilite_color = kwargs.pop('hilite', None)
        if 'linewidth' not in kwargs:
            kwargs['linewidth'] = self.linewidth
        p = lines.Line2D(x, y, **kwargs)
        p.highlight = highlight
        p.unhighlight = unhighlight
        return p

    def create_vertex(self, vertex, **kwargs):
        def highlight(p):
            lw = p.get_linewidth()
            c = p.get_color()
            p.unhilite = (c, lw)
            p.set_linewidth(2)
            if hilite_color:
                p.set_color(hilite_color)

        def unhighlight(p):
            c, lw = p.unhilite
            p.set_linewidth(lw)
            p.set_color(c)
            p.unhilite = None

        x = [vertex[0]]
        y = [vertex[1]]
        hilite_color = kwargs.pop('hilite', None)
        if 'linewidth' not in kwargs:
            kwargs['linewidth'] = self.linewidth
        p = lines.Line2D(x, y, **kwargs)
        p.highlight = highlight
        p.unhighlight = unhighlight
        return p

    def update_axis_limits(self, bbox):
        self.ax.set_xlim(bbox[0][0], bbox[1][0])
        self.ax.set_ylim(bbox[0][1], bbox[1][1])

    def fit_axis_limits(self):
        """ returns a numpy bounding box that fits the current data """
        pass

    def set_view_bbox(self, bbox):
        self.view_bbox = bbox
        self.update_axis_limits(bbox=self.view_bbox)

    def fit(self):
        self.set_view_bbox(self.fit_axis_limits())
        self.plot()

    def zoom(self, factor):
        bbox = self.view_bbox
        # calculate the bbox half-widths
        hlf_x, hlf_y = (bbox[1][0] - bbox[0][0])/2, (bbox[1][1] - bbox[0][1])/2
        # calculate the center of the bbox
        cen_x, cen_y = (bbox[1][0] + bbox[0][0])/2, (bbox[1][1] + bbox[0][1])/2
        # scale the bbox dimensions by the requested factor
        hlf_x *= factor
        hlf_y *= factor
        # rebuild the scaled bbox
        view_bbox = np.array([[cen_x-hlf_x, cen_y-hlf_y],
                              [cen_x+hlf_x, cen_y+hlf_y]])

        self.set_view_bbox(view_bbox)
        self.plot()

    def zoom_in(self):
        self.zoom(factor=0.8)

    def zoom_out(self):
        self.zoom(factor=1.2)

    def draw_frame(self, do_draw_frame):
        if do_draw_frame:
            self.ax.set_axis_on()
        else:
            self.ax.set_axis_off()
            self.tight_layout(pad=0.)

    def draw_axes(self, do_draw_axes):
        if do_draw_axes:
            self.ax.grid(True)
            self.ax.axvline(0, c=self._rgb['foreground'], lw=1)
            self.ax.axhline(0, c=self._rgb['foreground'], lw=1)
            if hasattr(self, 'header'):
                self.ax.set_title(self.header, pad=10.0, fontsize=18)
            if hasattr(self, 'x_label'):
                self.ax.set_xlabel(self.x_label)
            if hasattr(self, 'y_label'):
                self.ax.set_ylabel(self.y_label)
        else:
            self.ax.grid(False)

    def plot(self):
        try:
            self.ax.cla()
        except AttributeError:
            self.ax = self.add_subplot(1, 1, 1, aspect=self.aspect)

        for a in self.artists:
            a.set_picker(5)
            if isinstance(a, lines.Line2D):
                self.ax.add_line(a)
            elif isinstance(a, patches.Patch):
                self.ax.add_patch(a)
            else:
                self.ax.add_artist(a)

        if self.do_scale_bounds:
            self.view_bbox = util.scale_bounds(self.sys_bbox,
                                               self.oversize_factor)

        self.ax.set_aspect(self.aspect, adjustable='datalim')
        self.update_axis_limits(bbox=self.view_bbox)

        self.draw_frame(self.do_draw_frame)
        self.ax.set_facecolor(self._rgb['background1'])

        self.draw_axes(self.do_draw_axes)

        self.connect_events()
        self.canvas.draw_idle()

        return self

    def find_artists_at_location(self, event):
        artists = []
        for artist in self.ax.get_children():
            if hasattr(artist, 'shape'):
                inside, info = artist.contains(event)
                if inside:
                    shape, handle = artist.shape
                    artists.append(SelectInfo(artist, info))
                    if 'ind' in info:
                        logging.debug("on motion, artist {}: {}.{}, z={}, "
                                      "hits={}".format(len(artists),
                                      shape.get_label(), handle,
                                      artist.get_zorder(), info['ind']))
                    else:
                        logging.debug("on motion, artist {}: {}.{}, z={}"
                                      .format(len(artists), shape.get_label(),
                                              handle, artist.get_zorder()))

        return sorted(artists, key=lambda a: a.artist.get_zorder(),
                      reverse=True)

    def do_shape_action(self, event, target, event_key):
        if target is not None:
            shape, handle = target.artist.shape
            try:
                action = shape.actions[event_key]
                action(self, handle, event, target.info)
            except KeyError:
                pass

    def on_select(self, event):
        artist = event.artist
        display_artist_and_event('on_select', event, artist)

    def display_event(self, event):
        artist = self.hilited.artist if self.hilited else None
        display_artist_and_event('display_event', event, artist)

    def on_press(self, event):
        self.save_do_scale_bounds = self.do_scale_bounds
        self.do_scale_bounds = False
        target_artist = self.selected = self.hilited
        # if target_artist is not None:
        #     display_artist_and_event('on_press', event, self.selected.artist)
        self.do_action(event, target_artist, 'press')

    def on_motion(self, event):
        if self.selected is None:
            artists = self.find_artists_at_location(event)
            next_hilited = artists[0] if len(artists) > 0 else None
            # if next_hilited is not None:
            #     display_artist_and_event('on_motion ({})'.format(len(artists)),
            #                              event, next_hilited.artist)

            cur_art = self.hilited.artist if self.hilited is not None else None
            nxt_art = next_hilited.artist if next_hilited is not None else None
            if nxt_art is not cur_art:
                if self.hilited:
                    self.hilited.artist.unhighlight(self.hilited.artist)
                    self.hilited.artist.figure.canvas.draw()
                if next_hilited:
                    next_hilited.artist.highlight(next_hilited.artist)
                    next_hilited.artist.figure.canvas.draw()
                self.hilited = next_hilited
                if next_hilited is None:
                    logging.debug("hilite_change: no object found")
                else:
                    shape, handle = self.hilited.artist.shape
                    logging.debug("hilite_change: %s %s %d",
                                  shape.get_label(), handle,
                                  self.hilited.artist.get_zorder())
        else:
            # display_artist_and_event('on_drag', event, self.selected.artist)
            self.do_action(event, self.selected, 'drag')
            shape, handle = self.selected.artist.shape
            logging.debug("on_drag: %s %s %d", shape.get_label(), handle,
                          self.selected.artist.get_zorder())

    def on_release(self, event):
        'on release we reset the press data'
        logging.debug("on_release")
        # display_artist_and_event('on_release', event, self.selected.artist)

        self.do_action(event, self.selected, 'release')
        self.do_scale_bounds = self.save_do_scale_bounds
        self.selected = None
        self.action_complete()


class PanAction():
    ''' wrapper class to handle pan action, handing off to Axes '''

    def __init__(self, **kwargs):

        def on_press(fig, event):
            self._button_pressed = event.button
            fig.ax.start_pan(event.x, event.y, event.button)

        def on_drag(fig, event):
            fig.ax.drag_pan(self._button_pressed, event.key, event.x, event.y)
            fig.canvas.draw_idle()

        def on_release(fig, event):
            fig.ax.end_pan()
            # update figure view_bbox with result of pan action
            x_min, x_max = fig.ax.get_xbound()
            y_min, y_max = fig.ax.get_ybound()
            fig.view_bbox = np.array([[x_min, y_min], [x_max, y_max]])

        self.actions = {}
        self.actions['press'] = on_press
        self.actions['drag'] = on_drag
        self.actions['release'] = on_release


class ZoomBoxAction():
    """ handle zoom box action by using a RectangleSelector widget """

    def __init__(self, fig, **kwargs):
        def on_release(press_event, release_event):
            bbox = np.array([[press_event.xdata, press_event.ydata],
                             [release_event.xdata, release_event.ydata]])
            fig.set_view_bbox(bbox)
            fig.canvas.draw_idle()
            self.rubber_box.disconnect_events()
            fig.connect_events(self.saved_events)
            fig.action_complete()

        self.saved_events = fig.disconnect_events()
        rectprops = dict(edgecolor=fig._rgb['foreground'], fill=False)
        self.rubber_box = widgets.RectangleSelector(
            fig.ax, on_release, drawtype='box', useblit=False,
            button=[1, 3],  # don't use middle button
            minspanx=5, minspany=5, spancoords='pixels', rectprops=rectprops,
            interactive=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2017 Michael J. Hayford
""" Gap container class

.. Created on Fri Sep 15 17:06:17 2017

.. codeauthor: Michael J. Hayford
"""


from . import medium as m


class Gap:
    """ Gap container class.

    The gap class represents the space between 2 surfaces. It contains the
    media definition for the space and a (z) displacement between the
    adjacent surfaces.

    The most common use case is an optical system with surfaces centered on a
    common axis. The Gap structure implements this case in the simplest manner.
    More complicated transformations between surfaces are implemented using
    transformations associated with the surfaces themselves.
    """
    def __init__(self, t=0.0, med=m.Air()):
        self.thi = t
        self.medium = med

    def __repr__(self):
        return "Gap(t=%r, medium=%r)" % (self.thi, self.medium)

    def sync_to_restore(self, seq_model):
        if hasattr(self.medium, 'sync_to_restore'):
            self.medium.sync_to_restore()

    def apply_scale_factor(self, scale_factor):
        self.thi *= scale_factor



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2019 Michael J. Hayford
""" dict2D constructs a M x N dict using row_keys and col_keys as keys

.. Created on Wed Aug 14 21:08:56 2019

.. codeauthor: Michael J. Hayford
"""


def dict2D(row_keys, col_keys):
    """ returns a 2D dictionary with M row_keys and N col_keys """
    dict_2d = {}
    for r in row_keys:
        row = {}
        dict_2d[r] = row
        for c in col_keys:
            row[c] = {}
    return dict_2d


def row(dict_2d, row_key):
    """ returns a dict of the contents of row **row_key** of dict_2d """
    return dict_2d[row_key]


def col(dict_2d, col_key):
    """ returns a dict of the contents of column **col_key** of dict_2d """
    column = {}
    row_keys = dict_2d.keys()
    for r in row_keys:
        column[r] = dict_2d[r][col_key]
    return column
#    dict([(r, dict_2d[r][col_key]) for r in dict_2d.keys()])


def num_items_by_type(dict_2d, row_keys, col_keys):
    """ return a dict of the number of items in each row/col of dict_2d """
    len_items = dict([(rc, 0) for rc in row_keys+col_keys])
    for r in row_keys:
        for c in dict_2d[r]:
            num_items = len(dict_2d[r][c])
            len_items[r] += num_items
            len_items[c] += num_items
    return len_items


def num_items_by_cell(dict_2d, row_keys, col_keys):
    """ return a list of the number of items in each cell of dict_2d """
    num_items = []
    for r in row_keys:
        for c in dict_2d[r]:
            ni = len(dict_2d[r][c])
            num_items.append((r, c, ni))
    return num_items



""" package supplying Qt5 desktop application and associated functional support

    The ``rayoptics.qtgui`` subpackage provides a desktop app that runs under
    Anaconda. It also provides a series of higher level interfaces used by
    rayoptics. These include:

        - an interface that hosts matplotlib graphics
        - a table grid for numeric model displays (template-based)
        - docking panel support for python objects
        - iPython console window (desktop app only)
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Support for CODE V TLAs

.. codeauthor: Michael J. Hayford
"""
import csv
from pathlib import Path


class MapTLA:
    """ Create and maintain a dictionary of CODE V 3 letter commands """
    _d = {}

    def __init__(self):
        TLA, CmdFct, IndxQuals, DataType, Quals = range(5)
        if len(MapTLA._d) == 0:
            path = Path(__file__).resolve().parent
            with open(path / 'tla_mapping.csv', 'rU') as f:
                reader = csv.reader(f)
                for row in reader:
                    if row[TLA] is not '':
                        if row[Quals] is not '':
                            row[Quals] = row[Quals].split(',')
                        MapTLA._d[row[TLA]] = row[CmdFct:]

    def find(self, tla):
        try:
            return MapTLA._d[tla]
        except KeyError:
            return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Support for ray trace exception handling

.. Created on Wed Oct 24 15:22:40 2018

.. codeauthor: Michael J. Hayford
"""


class TraceError(Exception):
    """ Exception raised when ray tracing a model """


class TraceMissedSurfaceError(TraceError):
    """ Exception raised when ray misses a surface """
    def __init__(self, ifc=None, prev_seg=None):
        self.ifc = ifc
        self.prev_seg = prev_seg


class TraceTIRError(TraceError):
    """ Exception raised when ray TIRs on a surface """
    def __init__(self, inc_dir, normal, prev_indx, follow_indx):
        self.ifc = None
        self.int_pt = None
        self.inc_dir = inc_dir
        self.normal = normal
        self.prev_indx = prev_indx
        self.follow_indx = follow_indx


class TraceEvanescentRayError(TraceError):
    """ Exception raised when ray diffracts evanescently at a surface """
    def __init__(self, ifc, int_pt, inc_dir, normal, prev_indx, follow_indx):
        self.ifc = ifc
        self.int_pt = int_pt
        self.inc_dir = inc_dir
        self.normal = normal
        self.prev_indx = prev_indx
        self.follow_indx = follow_indx


class TraceRayBlockedError(TraceError):
    """ Exception raised when ray is blocked by an aperture on a surface """
    def __init__(self, ifc, int_pt, inc_dir, prev_indx, follow_indx):
        self.ifc = ifc
        self.int_pt = int_pt
        self.inc_dir = inc_dir
        self.prev_indx = prev_indx
        self.follow_indx = follow_indx



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" miscellaneous functions for working with numpy vectors and floats

.. Created on Wed May 23 15:27:06 2018

.. codeauthor: Michael J. Hayford
"""
import numpy as np
from numpy.linalg import norm
from math import sqrt, pi, acos


def normalize(v):
    """ return normalized version of input vector v """
    length = norm(v)
    if length == 0.0:
        return v
    else:
        return v/length


def distance_sqr_2d(pt0, pt1):
    """ return distance squared between 2d points pt0 and pt1 """
    return (pt0[0] - pt1[0])**2 + (pt0[1] - pt1[1])**2


def perpendicular_distance_2d(pt, pt1, pt2):
    """ return perpendicular distance of pt from the line between pt1 and pt2
    """
    return (((pt2[0] - pt1[0])*(pt1[1] - pt[1])
            - (pt1[0] - pt[0])*(pt2[1] - pt1[1]))
            / sqrt(distance_sqr_2d(pt2, pt1)))


def perpendicular_to_radial(pt, pt2):
    """ return perpendicular distance of pt from the line between the origin
    and pt2
    """
    return (pt[0]*pt2[1] - pt2[0]*pt[1]) / sqrt(pt2[0]**2 + pt2[1]**2)


def perpendicular_to_line(pt, pt1, pt2):
    """ return perpendicular distance of pt from the line between pt1 and pt2
    """
    d = perpendicular_distance_2d(pt, pt1, pt2)
    return (((pt2[0] - pt1[0])*(pt1[1] - pt[1])
            - (pt1[0] - pt[0])*(pt2[1] - pt1[1]))
            / sqrt(distance_sqr_2d(pt2, pt1)))


def perpendicular_from_origin(pt1, pt2):
    """ return perpendicular distance of the origin from the line between
    pt1 and pt2
    """
    return (((pt2[0] - pt1[0])*(pt1[1]) - (pt1[0])*(pt2[1] - pt1[1]))
            / sqrt(distance_sqr_2d(pt2, pt1)))


def projected_point_on_line(pt, pt1, pt2):
    e1 = pt2 - pt1
    e2 = pt - pt1
    # get dot product of e1, e2
    dot_prod = np.dot(e1, e2)
    # get squared length of e1
    len_sqr_e1 = e1[0]**2 + e1[1]**2
    p = np.array([(pt1[0] + (dot_prod * e1[0]) / len_sqr_e1),
                 (pt1[1] + (dot_prod * e1[1]) / len_sqr_e1)])
    return p


def projected_point_on_radial_line(pt, radial_pt):
    # get dot product of radial_pt, pt
    dot_prod = np.dot(radial_pt, pt)
    # get squared length of radial_pt
    len_sqr = radial_pt[0]**2 + radial_pt[1]**2
    p = np.array([(dot_prod * radial_pt[0]) / len_sqr,
                 (dot_prod * radial_pt[1]) / len_sqr])
    return p


def projected_point_on_radial_line_full(pt, radial_pt):
    x_prod = pt[0]*radial_pt[1] - radial_pt[0]*pt[1]
    # get dot product of radial_pt, pt
    dot_prod = np.dot(radial_pt, pt)
    # get squared length of radial_pt
    len_sqr = radial_pt[0]**2 + radial_pt[1]**2
    p = np.array([(dot_prod * radial_pt[0]) / len_sqr,
                 (dot_prod * radial_pt[1]) / len_sqr])
    dist = x_prod/sqrt(len_sqr)
    return p, dist


def euler2opt(e):
    """ convert right-handed euler angles to optical design convention,
        i.e. alpha and beta are left-handed
    """
    return np.array([-e[0], -e[1], e[2]])


def isanumber(a):
    """ returns true if input a can be converted to floating point number """
    try:
        float(a)
        bool_a = True
    except ValueError:
        bool_a = False

    return bool_a


def transpose(mat):
    """ transposes a m x n input list and returns the result """
    mat_t = []
    for j in range(len(mat[0])):
        mat_t.append([mat[i][j] for i in range(len(mat))])
    return mat_t


def circle_intersection_area(ra, rb, d):
    """ return the area of the intersection of 2 circles

    Args:
        ra: radius of first circle
        rb: radius of second circle
        d: separation of the circles' centers of curvature

    Returns:
        area of the circle intersection

    `Weisstein, Eric W. "Circle-Circle Intersection." From MathWorld--A Wolfram Web
    Resource. <http://mathworld.wolfram.com/Circle-CircleIntersection.html>`_
    """
    if ra < rb:  # sort into ascending order for convenience
        r, R = ra, rb
    else:
        r, R = rb, ra

    r2 = r**2
    if R >= r + d:  # smaller circle contained inside the larger one
        return pi*r2

    if d > r + R:  # circles are completely separated - no overlap
        return 0

    R2 = R**2
    d2 = d**2

    # calculate area via eq 14 in the referenced link
    p1 = r2*acos((d2 + r2 - R2)/(2*d*r))
    p2 = R2*acos((d2 + R2 - r2)/(2*d*R))
    p3 = sqrt((-d + r + R)*(d + r - R)*(d - r + R)*(d + r + R))/2
    area = p1 + p2 - p3
    return area


def compute_tangent_point_to_circle(CofC, r, pt):
    """ return the area of the intersection of 2 circles

    Args:
        CofC: center of curvature of circle (2d numpy array)
        r: radius of circle
        pt: 2d numpy array of point outside of circle

    Returns:
        the 2 tangent points for lines from pt to circle

    `gboffi. <https://math.stackexchange.com/users/467357/gboffi>`_ 
    `"How to find the equation of a line, tangent to a circle, that passes 
    through a given external point." StackExchange (version: 2019-05-30) 
    <https://math.stackexchange.com/a/3190374>`_
    """
    dxdy = pt - CofC
    dxdyr = np.array([-dxdy[1], dxdy[0]])
    d = sqrt(dxdy[0]**2 + dxdy[1]**2)
    if d > r:
        rho = r/d
        ad = rho**2
        bd = rho * sqrt(1 - rho**2)
        T1 = CofC + ad*dxdy + bd*dxdyr
        T2 = CofC + ad*dxdy - bd*dxdyr
        return T1, T2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Functions to support ray tracing a sequential optical model

.. Created on Thu Jan 25 11:01:04 2018

.. codeauthor: Michael J. Hayford
"""

import numpy as np
from numpy.linalg import norm
from math import sqrt, copysign

import rayoptics.optical.model_constants as mc
from rayoptics.optical.model_constants import Intfc, Gap, Indx, Tfrm, Zdir
from .traceerror import (TraceMissedSurfaceError, TraceTIRError,
                         TraceEvanescentRayError)


def bend(d_in, normal, n_in, n_out):
    """ refract incoming direction, d_in, about normal """
    try:
        normal_len = norm(normal)
        cosI = np.dot(d_in, normal)/normal_len
        sinI_sqr = 1.0 - cosI*cosI
        n_cosIp = copysign(sqrt(n_out*n_out - n_in*n_in*sinI_sqr), cosI)
        alpha = n_cosIp - n_in*cosI
        d_out = (n_in*d_in + alpha*normal)/n_out
        return d_out
    except ValueError:
        raise TraceTIRError(d_in, normal, n_in, n_out)


def reflect(d_in, normal):
    """ reflect incoming direction, d_in, about normal """
    normal_len = norm(normal)
    cosI = np.dot(d_in, normal)/normal_len
    d_out = d_in - 2.0*cosI*normal
    return d_out


def phase(ifc, inc_pt, d_in, normal, wvl, n_in, n_out):
    """ apply phase shift to incoming direction, d_in, about normal """
    try:
        d_out, dW = ifc.phase(inc_pt, d_in, normal, wvl)
        return d_out, dW
    except ValueError:
        raise TraceEvanescentRayError(ifc, inc_pt, d_in, normal, n_in, n_out)


def trace(seq_model, pt0, dir0, wvl, **kwargs):
    """ fundamental raytrace function

    Args:
        seq_model: the sequential model to be traced
        pt0: starting point in coords of first interface
        dir0: starting direction cosines in coords of first interface
        wvl: wavelength in nm
        eps: accuracy tolerance for surface intersection calculation

    Returns:
        (**ray**, **op_delta**, **wvl**)

        - **ray** is a list for each interface in **path_pkg** of these
          elements: [pt, after_dir, after_dst, normal]

            - pt: the intersection point of the ray
            - after_dir: the ray direction cosine following the interface
            - after_dst: after_dst: the geometric distance to the next
              interface
            - normal: the surface normal at the intersection point

        - **op_delta** - optical path wrt equally inclined chords to the
          optical axis
        - **wvl** - wavelength (in nm) that the ray was traced in
    """
    path = seq_model.path(wvl)
    return trace_raw(path, pt0, dir0, wvl, **kwargs)


def trace_raw(path, pt0, dir0, wvl, eps=1.0e-12):
    """ fundamental raytrace function

    Args:
        path: an iterator containing interfaces and gaps to be traced.
              for each iteration, the sequence or generator should return a
              list containing: **Intfc, Gap, Trfm, Index, Z_Dir**
        pt0: starting point in coords of first interface
        dir0: starting direction cosines in coords of first interface
        wvl: wavelength in nm
        eps: accuracy tolerance for surface intersection calculation

    Returns:
        (**ray**, **op_delta**, **wvl**)

        - **ray** is a list for each interface in **path** of these
          elements: [pt, after_dir, after_dst, normal]

            - pt: the intersection point of the ray
            - after_dir: the ray direction cosine following the interface
            - after_dst: after_dst: the geometric distance to the next
              interface
            - normal: the surface normal at the intersection point

        - **op_delta** - optical path wrt equally inclined chords to the
          optical axis
        - **wvl** - wavelength (in nm) that the ray was traced in
    """
    ray = []
    eic = []

    # trace object surface
    obj = next(path)
    srf_obj = obj[Intfc]
    dst_b4, pt_obj = srf_obj.intersect(pt0, dir0)

    before = obj
    before_pt = pt_obj
    before_dir = dir0
    before_normal = srf_obj.normal(before_pt)
    tfrm_from_before = before[Tfrm]
    z_dir_before = before[Zdir]

    op_delta = 0.0
    surf = 0
    # loop of remaining surfaces in path
    while True:
        try:
            after = next(path)

            rt, t = tfrm_from_before
            b4_pt, b4_dir = rt.dot(before_pt - t), rt.dot(before_dir)

            pp_dst = -b4_pt.dot(b4_dir)
            pp_pt_before = b4_pt + pp_dst*b4_dir

            ifc = after[Intfc]
            z_dir_after = after[Zdir]

            # intersect ray with profile
            pp_dst_intrsct, inc_pt = ifc.intersect(pp_pt_before, b4_dir,
                                                   eps=eps, z_dir=z_dir_before)
            dst_b4 = pp_dst + pp_dst_intrsct
            ray.append([before_pt, before_dir, dst_b4, before_normal])

            normal = ifc.normal(inc_pt)

            eic_dst_before = ((inc_pt.dot(b4_dir) + z_dir_before*inc_pt[2]) /
                              (1.0 + z_dir_before*b4_dir[2]))

            # if the interface has a phase element, process that first
            if hasattr(ifc, 'phase_element'):
                doe_dir, phs = phase(ifc, inc_pt, b4_dir, normal, wvl,
                                     before[Indx], after[Indx])
                # the output of the phase element becomes the input for the
                #  refraction/reflection calculation
                b4_dir = doe_dir
                op_delta += phs

            # refract or reflect ray at interface
            if ifc.interact_mode == 'reflect':
                after_dir = reflect(b4_dir, normal)
            elif ifc.interact_mode == 'transmit':
                after_dir = bend(b4_dir, normal, before[Indx], after[Indx])
            else:  # no action, input becomes output
                after_dir = b4_dir

            eic_dst_after = ((inc_pt.dot(after_dir) + z_dir_after*inc_pt[2]) /
                             (1.0 + z_dir_after*after_dir[2]))

            surf += 1

            # Per `Hopkins, 1981 <https://dx.doi.org/10.1080/713820605>`_, the
            #  propagation direction is given by the direction cosines of the
            #  ray and therefore doesn't require the use of a negated
            #  refractive index following a reflection. Thus we use the
            #  (positive) refractive indices from the seq_model.rndx array.
            dW = after[Indx]*eic_dst_after - before[Indx]*eic_dst_before
            eic.append([before[Indx], eic_dst_before,
                        after[Indx], eic_dst_after, dW])

#            print("after:", surf, inc_pt, after_dir)
#            print("e{}= {:12.5g} e{}'= {:12.5g} dW={:10.8g} n={:8.5g}"
#                  " n'={:8.5g} zdirb4={:2.0f} zdiraftr={:2.0f}"
#                  .format(surf, eic_dst_before, surf, eic_dst_after, dW,
#                          before[Indx], after[Indx], z_dir_before, z_dir_after))
            before_pt = inc_pt
            before_normal = normal
            before_dir = after_dir
            z_dir_before = z_dir_after
            before = after
            tfrm_from_before = before[Tfrm]

        except TraceMissedSurfaceError as ray_miss:
            ray.append([before_pt, before_dir, pp_dst, before_normal])
            ray_miss.surf = surf+1
            ray_miss.ifc = ifc
            ray_miss.prev_tfrm = before[Tfrm]
            ray_miss.ray = ray
            raise ray_miss

        except TraceTIRError as ray_tir:
            ray.append([inc_pt, before_dir, 0.0, normal])
            ray_tir.surf = surf+1
            ray_tir.ifc = ifc
            ray_tir.int_pt = inc_pt
            ray_tir.ray = ray
            raise ray_tir

        except TraceEvanescentRayError as ray_evn:
            ray.append([inc_pt, before_dir, 0.0, normal])
            ray_evn.surf = surf+1
            ray_evn.ifc = ifc
            ray_evn.int_pt = inc_pt
            ray_evn.ray = ray
            raise ray_evn

        except StopIteration:
            ray.append([inc_pt, after_dir, 0.0, normal])
            if len(eic) > 1:
                P, P1k, Ps = calc_path_length(eic, offset=1)
                op_delta += P
            break

    return ray, op_delta, wvl


def calc_path_length(eic, offset=0):
    """ given eic array, compute path length between outer surfaces

    Args:
        eic: equally inclined chord array
        offset (int): beginning index of eic array wrt the object interface

    Returns:
        float: path length
    """
    P1k = -eic[1-offset][2]*eic[1-offset][3] + eic[-2][0]*eic[-2][1]
    Ps = 0.
    for i in range(2-offset, len(eic)-2):
        Ps -= eic[i][4]
#        Ps -= eic[i][2]*eic[i][3] - eic[i][0]*eic[i][1]
    P = P1k + Ps
    return P, P1k, Ps


def eic_distance(r, r0):
    """ calculate equally inclined chord distance between 2 rays

    Args:
        r: (p, d), where p is a point on the ray r and d is the direction
           cosine of r
        r0: (p0, d0), where p0 is a point on the ray r0 and d0 is the direction
            cosine of r0

    Returns:
        float: distance along r from equally inclined chord point to p
    """
    # eq 3.9
    e = (np.dot(r[1] + r0[1], r[0] - r0[0]) / (1. + np.dot(r[1], r0[1])))
    return e


def wave_abr(fld, wvl, foc, ray_pkg):
    """ computes optical path difference (OPD) for ray_pkg at fld and wvl

    The main references for the calculations are in the H. H. Hopkins paper
    `Calculation of the Aberrations and Image Assessment for a General Optical
    System <https://doi.org/10.1080/713820605>`_

    Args:
        fld: :class:`~.Field` point for wave aberration calculation
        wvl: wavelength of ray
        foc: :class:`~.FocusRange` instance, to specify defocus
        ray_pkg: input tuple of ray, ray_op, wvl

    Returns:
        (**opd**, **e1**, **ekp**, **ep**)

        - **opd** - OPD of ray wrt chief ray at **fld**
        - **e1** - eic in object space, prior to first interface
        - **ekp** - eic in image space, following final interface
        - **ep** - eic to reference sphere intersection
    """
    return wave_abr_real_coord(fld, wvl, foc, ray_pkg)
#    return wave_abr_HHH(fld.ref_sphere_pkg, fld.chief_ray_pkg, ray_pkg)


def wave_abr_real_coord(fld, wvl, foc, ray_pkg):
    ref_sphere, parax_data, n_obj, n_img, z_dir = fld.ref_sphere
    image_pt, cr_exp_pt, cr_exp_dist, ref_dir, ref_sphere_radius = ref_sphere
    chief_ray, chief_ray_op, wvl = fld.chief_ray[0]
    ray, ray_op, wvl = ray_pkg
    fod = parax_data[2]
    k = -2  # last interface in sequence

    # eq 3.12
    e1 = eic_distance((ray[1][mc.p], ray[0][mc.d]),
                      (chief_ray[1][mc.p], chief_ray[0][mc.d]))
    # eq 3.13
    ekp = eic_distance((ray[k][mc.p], ray[k][mc.d]),
                       (chief_ray[k][mc.p], chief_ray[k][mc.d]))

    dst = ekp - cr_exp_dist

    eic_exp_pt = ray[k][mc.p] - dst*ray[k][mc.d]
#    eic_exp_pt[2] -= cr_exp_dist
    p_coord = eic_exp_pt - cr_exp_pt
    F = ref_dir.dot(ray[k][mc.d]) - ray[k][mc.d].dot(p_coord)/ref_sphere_radius
    J = p_coord.dot(p_coord)/ref_sphere_radius - 2.0*ref_dir.dot(p_coord)
    ep = J/(F + sqrt(F**2 + J/ref_sphere_radius))

    opd = -n_obj*e1 - ray_op + n_img*ekp + chief_ray_op - n_img*ep
    return opd, e1, ekp, ep


def wave_abr_HHH(fld, wvl, foc, ray_pkg):
    ref_sphere, parax_data, n_obj, n_img, z_dir = fld.ref_sphere
    image_pt, cr_exp_pt, ref_dir, ref_sphere_radius = ref_sphere
    chief_ray, chief_ray_op, wvl = fld.chief_ray
    ray, ray_op, wvl = ray_pkg
    ax_ray, pr_ray, fod = parax_data
    k = -2  # last interface in sequential model
    ax_k = ax_ray[k]
    pr_k = pr_ray[k]
    ht, slp = range(2)
    H = n_img*(pr_k[ht]*ax_k[slp] - ax_k[ht]*pr_k[slp])

    # eq 3.12
    e1 = eic_distance((ray[1][mc.p], ray[0][mc.d]),
                      (chief_ray[1][mc.p], chief_ray[0][mc.d]))
    # eq 3.13
    ekp = eic_distance((ray[k][mc.p], ray[k][mc.d]),
                       (chief_ray[k][mc.p], chief_ray[k][mc.d]))

    # eq 4.33
    eic_pt = ray[k][mc.p] - ekp*ray[k][mc.d]
    print("eic_pt", eic_pt)

    Nk_cr = chief_ray[k][mc.d][2]
    Zk_cr = chief_ray[k][mc.p][2]

    def reduced_pupil_coord(X, L, e):
        coef1 = -n_img/(H * Nk_cr)
        xp = coef1*(pr_k[slp]*(Nk_cr*(X - L*e) - L*Zk_cr) +
                    pr_k[ht]*L)
        return xp

    # eq 5.4
    xp_ray = reduced_pupil_coord(ray[k][mc.p][0], ray[k][mc.d][0], ekp)
    yp_ray = reduced_pupil_coord(ray[k][mc.p][1], ray[k][mc.d][1], ekp)
    # eq 5.5
    xp_cr = reduced_pupil_coord(chief_ray[k][mc.p][0], chief_ray[k][mc.d][0], 0.)
    yp_cr = reduced_pupil_coord(chief_ray[k][mc.p][1], chief_ray[k][mc.d][1], 0.)
    # eq 5.6
    zp_ray = -(((ray[k][mc.d][0] + chief_ray[k][mc.d][0])*(xp_ray - xp_cr) +
                (ray[k][mc.d][1] + chief_ray[k][mc.d][1])*(yp_ray - yp_cr)) /
                (ray[k][mc.d][2] + chief_ray[k][mc.d][2]))

    rpc_ray = np.array([xp_ray, yp_ray, zp_ray])
    rpc_cr = np.array([xp_cr, yp_cr, 0.])
    print("rpc ray", xp_ray, yp_ray, zp_ray)
    print("rpc cr", xp_cr, yp_cr, 0.)
    # eq 4.11
    G0_ref = n_img*ax_k[slp]*image_pt[0]
    H0_ref = n_img*ax_k[slp]*image_pt[1]

    def reduced_image_coord(X, L, Z, N):
        G0 = (n_img/N)*(ax_k[slp]*(N*X - L*Z) + ax_k[ht]*L)
        return G0

    # eq 5.13
    G0_ray = reduced_image_coord(ray[k][mc.p][0], ray[k][mc.d][0],
                                 ray[k][mc.p][2], ray[k][mc.d][2])
    H0_ray = reduced_image_coord(ray[k][mc.p][1], ray[k][mc.d][1],
                                 ray[k][mc.p][2], ray[k][mc.d][2])
    # eq 5.14
    G0_cr = reduced_image_coord(chief_ray[k][mc.p][0], chief_ray[k][mc.d][0],
                                Zk_cr, Nk_cr)
    H0_cr = reduced_image_coord(chief_ray[k][mc.p][1], chief_ray[k][mc.d][1],
                                Zk_cr, Nk_cr)
    print("G0, H0_ref; G0, H0_cr:", G0_ref, H0_ref, G0_cr, H0_cr)
    # eq 4.17
    a = pr_k[slp]*G0_ref/H + ax_k[slp]*xp_cr
    b = pr_k[slp]*H0_ref/H + ax_k[slp]*yp_cr
    g = z_dir/sqrt(1. - a**2 + b**2)
    # eq 4.18
    ref_dir = np.array([-a*g, -b*g, g])
    print("ref_dir, cr_dir", ref_dir, chief_ray[k][mc.d])
    # eq 4.25
    F = (np.dot(ref_dir, ray[k][mc.d]) + ref_dir[2]*ax_k[slp] *
         np.dot(chief_ray[k][mc.d], (rpc_ray - rpc_cr)))

    # eq 4.28
    Ja = (ref_dir[2]*ax_k[slp] *
          np.dot((eic_pt - chief_ray[k][0]), (rpc_ray - rpc_cr)))
    Jb = -(2.0*(ref_dir[2]/Nk_cr)*(ax_k[ht] - ax_k[slp]*Zk_cr) *
           np.dot(chief_ray[k][mc.d], (rpc_ray - rpc_cr)))
    Jc = (2.0*(ref_dir[2]/n_img)*((G0_cr - G0_ref)*(xp_ray - xp_cr) +
                                  (H0_cr - H0_ref)*(yp_ray - yp_cr)))
    J = Ja + Jb + Jc
    print("F, J, Ja, Jb, Jc", F, J, Ja, Jb, Jc)
#    J = ((ref_dir[2]*ax_k[slp] *
#         np.dot((eic_pt - chief_ray[k][0]), (rpc_ray - rpc_cr))) -
#         2.0*(ref_dir[2]/Nk_cr)*(ax_k[ht] - ax_k[slp]*Zk_cr) *
#         np.dot(chief_ray[k][mc.d], (rpc_ray - rpc_cr)) +
#         2.0*(ref_dir[2]/n_img)*((G0_cr - G0_ref)*(xp_ray - xp_cr) +
#                                 (H0_cr - H0_ref)*(yp_ray - yp_cr)))

#    # eq 4.29 Q' = image_pt
#    F = (np.dot(chief_ray[k][mc.d], ray[k][mc.d]) + Nk_cr*ax_k[slp] *
#         np.dot(chief_ray[k][mc.d], (rpc_ray - rpc_cr)))
#    J = ((Nk_cr*ax_k[slp] *
#         np.dot((eic_pt - chief_ray[k][0]), (rpc_ray - rpc_cr))) -
#         2.0*(ax_k[ht] - ax_k[slp]*Zk_cr) *
#         np.dot(chief_ray[k][mc.d], (rpc_ray - rpc_cr)))

    # eq 4.21
    ep = J/(F + sqrt(F**2 + J*(n_img*ax_k[slp]*pr_k[slp]*ref_dir[2] / H)))

    print("F, J, ep (canon)", F, J, ep)

    # eq 3.14/3.22, using 3.15
    opd = -n_obj*e1 - ray_op + n_img*ekp + chief_ray_op - n_img*ep
    return opd, e1, ekp, ep


def transfer_to_exit_pupil(interface, ray_seg, exp_dst_parax):
    if interface.decenter:
        # get transformation info after surf
        r, t = interface.decenter.tform_after_surf()
        rt = r.transpose()
        b4_pt, b4_dir = rt.dot(ray_seg[0] - t), rt.dot(ray_seg[1])
    else:
        b4_pt, b4_dir = ray_seg[0], ray_seg[1]

    h = b4_pt[0]**2 + b4_pt[1]**2
    u = b4_dir[0]**2 + b4_dir[1]**2
    if u == 0.0:
        dst = exp_dst_parax
    else:
        dst = -sqrt(h/u)

    exp_pt = b4_pt + dst*b4_dir

    return exp_pt, b4_dir, dst


def eic_path_accumulation(ray, rndx, lcl_tfrms, z_dir):
    """ computes equally inclined chords and path info for ray

    Args:
        ray: ray data for traced ray
        rndx: refractive index array
        lcl_tfrms: local surface interface transformation data
        z_dir: z direction array

    Returns:
        (**eic**, **op_delta**)

        - **eic** - list of [n_before, eic_dst_before, n_after, eic_dst_after,
          dW]
        - **op_delta** - optical path wrt equally inclined chords to the
          optical axis
    """
    eic = []

    z_dir_before = z_dir[0]
    n_before = rndx[0]

    before_dir = ray[0][1]

    for i, r in enumerate(ray):
        rotT, _ = lcl_tfrms[i]
        b4_dir = rotT.dot(before_dir)

        z_dir_after = z_dir[i]

        inc_pt = ray[i][0]
        eic_dst_before = ((inc_pt.dot(b4_dir) + z_dir_before*inc_pt[2]) /
                          (1.0 + z_dir_before*b4_dir[2]))

        after_dir = ray[i][1]
        eic_dst_after = ((inc_pt.dot(after_dir) + z_dir_after*inc_pt[2]) /
                         (1.0 + z_dir_after*after_dir[2]))

        # Per `Hopkins, 1981 <https://dx.doi.org/10.1080/713820605>`_, the
        #  propagation direction is given by the direction cosines of the ray
        #  and therefore doesn't require the use of a negated refractive index
        #  following a reflection. Thus we use the (positive) refractive indices
        #  from the seq_model.rndx array.
        n_after = rndx[i]
        dW = n_after*eic_dst_after - n_before*eic_dst_before

        eic.append([n_before, eic_dst_before, n_after, eic_dst_after, dW])
        print("e{}= {:12.5g} e{}'= {:12.5g} dW={:10.8g} n={:8.5g}"
              " n'={:8.5g}".format(i, eic_dst_before,
                                   i, eic_dst_after,
                                   dW, n_before, n_after))

        n_before = n_after
        before_dir = after_dir
        z_dir_before = z_dir_after

    P, P1k, Ps = calc_path_length(eic)
    return eic, P



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Ray Optics GUI Application

Relies on PyQt5

.. Created on Mon Feb 12 09:24:01 2018

.. codeauthor: Michael J. Hayford
"""

import sys
import logging
from pathlib import Path

from PyQt5.QtCore import Qt as qt
from PyQt5.QtCore import QEvent
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import (QApplication, QAction, QMainWindow, QMdiArea,
                             QFileDialog, QTableView, QWidget, QMenu,
                             QVBoxLayout, QGraphicsView, QGraphicsScene)
from PyQt5.QtCore import pyqtSlot
import qdarkstyle
from qdarkstyle.palette import DarkPalette

from traitlets.config.configurable import MultipleInstanceError

from rayoptics.optical.trace import RaySeg
import rayoptics.gui.appcmds as cmds
from rayoptics.gui.appmanager import ModelInfo, AppManager
import rayoptics.qtgui.dockpanels as dock
from rayoptics.qtgui.graphicsitems import OpticalElement, RayBundle
from rayoptics.qtgui.ipyconsole import create_ipython_console

from rayoptics.optical import firstorder
from rayoptics.optical import trace


class MainWindow(QMainWindow):
    count = 0

    def __init__(self, parent=None, qtapp=None):
        super().__init__(parent)
        self.qtapp = qtapp
        self.mdi = QMdiArea()
        self.setCentralWidget(self.mdi)

        self.app_manager = AppManager(None, gui_parent=self)
        self.mdi.subWindowActivated.connect(self.app_manager.
                                            on_view_activated)

        self.is_dark = self.light_or_dark(False)

        self.left = 100
        self.top = 50
        self.width = 1800
        self.height = 1200
        self.setGeometry(self.left, self.top, self.width, self.height)

        bar = self.menuBar()

        file_menu = bar.addMenu("File")
        file_menu.addAction("New")
        file_menu.addAction("Open...")
        file_menu.addSeparator()
        file_menu.addAction("Save")
        file_menu.addAction("Save As...")
        file_menu.addAction("Close")
        file_menu.triggered[QAction].connect(self.file_action)

        view_menu = bar.addMenu("View")
        view_menu.addAction("Spec Sheet")
        view_menu.addAction("Optical Layout")
        view_menu.addAction("Lens Table")
        view_menu.addAction("Element Table")
        view_menu.addAction("Lens View")
        view_menu.triggered[QAction].connect(self.view_action)

        parax_menu = bar.addMenu("Paraxial Model")
        parax_menu.addAction("Paraxial Model")
        parax_menu.addAction("y-ybar View")
        parax_menu.addAction("nu-nubar View")
        parax_menu.addAction("yui Ray Table")
        parax_menu.addAction("3rd Order Aberrations")
        parax_menu.triggered[QAction].connect(self.view_action)

        analysis_menu = bar.addMenu("Analysis")
        view_menu.addAction("Ray Table")
        analysis_menu.addAction("Ray Fans")
        analysis_menu.addAction("OPD Fans")
        analysis_menu.addAction("Spot Diagram")
        analysis_menu.addAction("Wavefront Map")
        analysis_menu.addAction("Astigmatism Curves")
        analysis_menu.triggered[QAction].connect(self.view_action)

        wnd_menu = bar.addMenu("Window")
        wnd_menu.addAction("Cascade")
        wnd_menu.addAction("Tiled")
        wnd_menu.addSeparator()
        wnd_menu.addAction("Light UI")
        wnd_menu.addAction("Dark UI")
        wnd_menu.addSeparator()

        dock.create_dock_windows(self)
        for pi in dock.panels.values():
            wnd_menu.addAction(pi.menu_action)

        wnd_menu.triggered[QAction].connect(self.window_action)

        self.setWindowTitle("Ray Optics")
        self.show()

        self.new_model()
        self.add_ipython_subwindow()

#        pth = Path(__file__).resolve()
#        try:
#            root_pos = pth.parts.index('rayoptics')
#        except ValueError:
#            logging.debug("Can't find rayoptics: path is %s", pth)
#        else:
#            path = Path(*pth.parts[:root_pos+1])
#            self.open_file(path / "codev/tests/asp46.seq")
#            self.open_file(path / "codev/tests/dar_test.seq")
#            self.open_file(path / "codev/tests/paraboloid.seq")
#            self.open_file(path / "codev/tests/paraboloid_f8.seq")
#            self.open_file(path / "codev/tests/schmidt.seq")
#            self.open_file(path / "codev/tests/questar35.seq")
#            self.open_file(path / "codev/tests/rc_f16.seq")
#            self.open_file(path / "codev/tests/ag_dblgauss.seq")
#            self.open_file(path / "codev/tests/threemir.seq")
#            self.open_file(path / "codev/tests/folded_lenses.seq")
#            self.open_file(path / "codev/tests/unfolded_lenses_w_ape.seq")
#            self.open_file(path / "codev/tests/lens_reflection_test.seq")
#            self.open_file(path / "codev/tests/dec_tilt_test.seq")
#            self.open_file(path / "codev/tests/landscape_lens.seq")
#            self.open_file(path / "codev/tests/mangin.seq")
#            self.open_file(path / "optical/tests/cell_phone_camera.roa")
#            self.open_file(path / "optical/tests/singlet_f3.roa")

#        try:
#            root_pos = pth.parts.index('ray-optics')
#        except ValueError:
#            logging.debug("Can't find ray-optics: path is %s", pth)
#        else:
#            path = Path(*pth.parts[:root_pos+1])
#            self.open_file(path / "models/TwoMirror.roa")
#            self.open_file(path / "models/TwoSphericalMirror.roa")
#            self.open_file(path / "models/Sasian Triplet.roa")
#            self.open_file(path / "models/singlet_f5.roa")
#            self.open_file(path / "models/thinlens.roa")
#            self.open_file(path / "models/thin_triplet.roa")
#            self.open_file(path / "models/Cassegrain.roa")
#            self.open_file(path / "models/Ritchey_Chretien.roa")
#        finally:
#            self.add_ipython_subwindow()

    def add_subwindow(self, widget, model_info):
        sub_wind = self.mdi.addSubWindow(widget)
        self.app_manager.add_view(sub_wind, widget, model_info)
        MainWindow.count += 1
        return sub_wind

    def delete_subwindow(self, sub_wind):
        self.app_manager.delete_view(sub_wind)
        self.mdi.removeSubWindow(sub_wind)
        MainWindow.count -= 1

    def add_ipython_subwindow(self):
        try:
            create_ipython_console(self, 'iPython console', 600, 400)
        except MultipleInstanceError:
            logging.debug("Unable to open iPython console. "
                          "MultipleInstanceError")
        except Exception as inst:
            print(type(inst))    # the exception instance
            print(inst.args)     # arguments stored in .args
            print(inst)          # __str__ allows args to be printed directly,
            pass                 # but may be overridden in exception subclasses

    def initial_window_offset(self):
        offset_x = 50
        offset_y = 25
        orig_x = (MainWindow.count - 1)*offset_x
        orig_y = (MainWindow.count - 1)*offset_y
        return orig_x, orig_y

    def file_action(self, q):
        if q.text() == "New":
            self.new_model()

        if q.text() == "Open...":
            options = QFileDialog.Options()
            # options |= QFileDialog.DontUseNativeDialog
            fileName, _ = QFileDialog.getOpenFileName(
                          self,
                          "QFileDialog.getOpenFileName()",
                          "",
                          "CODE V Files (*.seq);;Ray-Optics Files (*.roa)",
                          options=options)
            if fileName:
                logging.debug("open file: %s", fileName)
                self.open_file(fileName)

        if q.text() == "Save As...":
            options = QFileDialog.Options()
            # options |= QFileDialog.DontUseNativeDialog
            fileName, _ = QFileDialog.getSaveFileName(
                         self,
                          "QFileDialog.getSaveFileName()",
                          "",
                          "Ray-Optics Files (*.roa);;All Files (*)",
                          options=options)
            if fileName:
                logging.debug("save file: %s", fileName)
                self.save_file(fileName)

        if q.text() == "Close":
            self.close_model()

    def new_model(self):
        iid = cmds.create_new_ideal_imager(gui_parent=self,
                                           conjugate_type='infinite')

        self.refresh_app_ui()

    def open_file(self, file_name):
        self.cur_filename = file_name
        self.app_manager.set_model(cmds.open_model(file_name))
        self.is_changed = True
        self.create_lens_table()
        cmds.create_live_layout_view(self.app_manager.model, gui_parent=self)
#        cmds.create_lens_layout_view(self.app_manager.model, gui_parent=self)
#        self.create_2D_lens_view()
        self.refresh_app_ui()

    def save_file(self, file_name):
        self.app_manager.model.save_model(file_name)
        self.cur_filename = file_name
        self.is_changed = False

    def close_model(self):
        """ NOTE: this does not check to save a modified model """
        self.app_manager.close_model(self.delete_subwindow)

    def view_action(self, q):
        opt_model = self.app_manager.model

        if q.text() == "Spec Sheet":
            cmds.create_new_ideal_imager(opt_model=opt_model, gui_parent=self)

        if q.text() == "Optical Layout":
            cmds.create_live_layout_view(opt_model, gui_parent=self)

        if q.text() == "Lens View":
            cmds.create_lens_layout_view(opt_model, gui_parent=self)
#            self.create_2D_lens_view()

        if q.text() == "Lens Table":
            self.create_lens_table()

        if q.text() == "Element Table":
            model = cmds.create_element_table_model(opt_model)
            self.create_table_view(model, "Element Table")

        if q.text() == "Ray Fans":
            cmds.create_ray_fan_view(opt_model, "Ray", gui_parent=self)

        if q.text() == "OPD Fans":
            cmds.create_ray_fan_view(opt_model, "OPD", gui_parent=self)

        if q.text() == "Spot Diagram":
            cmds.create_ray_grid_view(opt_model, gui_parent=self)

        if q.text() == "Wavefront Map":
            cmds.create_wavefront_view(opt_model, gui_parent=self)

        if q.text() == "Astigmatism Curves":
            cmds.create_field_curves(opt_model, gui_parent=self)

        if q.text() == "3rd Order Aberrations":
            cmds.create_3rd_order_bar_chart(opt_model, gui_parent=self)

        if q.text() == "y-ybar View":
            cmds.create_paraxial_design_view_v2(opt_model, 'ht',
                                                gui_parent=self)

        if q.text() == "nu-nubar View":
            cmds.create_paraxial_design_view_v2(opt_model, 'slp',
                                                gui_parent=self)

        if q.text() == "yui Ray Table":
            model = cmds.create_parax_table_model(opt_model)
            self.create_table_view(model, "Paraxial Ray Table")

        if q.text() == "Paraxial Model":
            model = cmds.create_parax_model_table(opt_model)
            self.create_table_view(model, "Paraxial Model")

        if q.text() == "Ray Table":
            self.create_ray_table(opt_model)

    def window_action(self, q):
        if q.text() == "Cascade":
            self.mdi.cascadeSubWindows()

        if q.text() == "Tiled":
            self.mdi.tileSubWindows()

        if q.text() == "Light UI":
            self.is_dark = self.light_or_dark(False)
            self.app_manager.sync_light_or_dark(self.is_dark)

        if q.text() == "Dark UI":
            self.is_dark = self.light_or_dark(True)
            self.app_manager.sync_light_or_dark(self.is_dark)

    def light_or_dark(self, is_dark):
        """ set the UI to a light or dark scheme.
        
        Qt doesn't seem to support controlling the MdiArea's background from a
        style sheet. Set the widget directly and save the original color
        to reset defaults.
        """
        if not hasattr(self, 'mdi_background'):
            self.mdi_background = self.mdi.background()

        if is_dark:
            self.qtapp.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
            rgb = DarkPalette.color_palette()
            self.mdi.setBackground(QColor(rgb['COLOR_BACKGROUND_NORMAL']))
        else:
            self.qtapp.setStyleSheet('')
            self.mdi.setBackground(self.mdi_background)
        return is_dark

    def create_lens_table(self):
        seq_model = self.app_manager.model.seq_model

        def set_stop_surface(stop_surface):
            seq_model.stop_surface = stop_surface
            self.refresh_gui()

        def handle_context_menu(point):
            row = vheader.logicalIndexAt(point.y())
            # show menu about the row
            menu = QMenu(self)
            if row != seq_model.stop_surface:
                menu.addAction('Set Stop Surface',
                               lambda: set_stop_surface(row))
            if seq_model.stop_surface is not None:
                menu.addAction('Float Stop Surface',
                               lambda: set_stop_surface(None))
            menu.popup(vheader.mapToGlobal(point))

        model = cmds.create_lens_table_model(seq_model)
        view = self.create_table_view(model, "Surface Data Table")
        vheader = view.verticalHeader()
        vheader.setContextMenuPolicy(qt.CustomContextMenu)
        vheader.customContextMenuRequested.connect(handle_context_menu)

    def create_ray_table(self, opt_model):
        osp = opt_model.optical_spec
        pupil = [0., 1.]
        fi = 0
        wl = osp.spectral_region.reference_wvl
        fld, wvl, foc = osp.lookup_fld_wvl_focus(fi, wl)
        ray, ray_op, wvl = trace.trace_base(opt_model, pupil, fld, wvl)
#        ray, ray_op, wvl, opd = trace.trace_with_opd(opt_model, pupil,
#                                                     fld, wvl, foc)

#        cr = trace.RayPkg(ray, ray_op, wvl)
#        s, t = trace.trace_coddington_fan(opt_model, cr, foc)

        ray = [RaySeg(*rs) for rs in ray]
        model = cmds.create_ray_table_model(opt_model, ray)
        self.create_table_view(model, "Ray Table")

    def create_2D_lens_view(self):
        scene2d = QGraphicsScene()
        self.create_element_model(scene2d)
        self.create_ray_model(scene2d)
        scene2d.setBackgroundBrush(QColor(237, 243, 254))  # light blue
        sceneRect2d = scene2d.sceneRect()

        # construct the top level widget
        widget = QWidget()
        # construct the top level layout
        layout = QVBoxLayout(widget)

        # set the layout on the widget
        widget.setLayout(layout)

        sub = self.add_subwindow(widget, ModelInfo(self.app_manager.model,
                                         MainWindow.update_2D_lens_view,
                                         (scene2d,)))
        sub.setWindowTitle("2D Lens View")
        view_width = 660
        view_ht = 440
        view_ratio = view_width/view_ht
        orig_x, orig_y = self.initial_window_offset()
        sub.setGeometry(orig_x, orig_y, view_width, view_ht)

        self.gview2d = QGraphicsView(scene2d)
        scene_ratio = sceneRect2d.width()/sceneRect2d.height()
        oversize_fraction = 1.2
        if scene_ratio > view_ratio:
            view_scale = view_width/(oversize_fraction*sceneRect2d.width())
        else:
            view_scale = view_ht/(oversize_fraction*sceneRect2d.height())

        self.gview2d.scale(view_scale, view_scale)
        layout.addWidget(self.gview2d)

        sub.show()

        return sub

    def update_2D_lens_view(scene2d):
        for gi in scene2d.items():
            gi.prepareGeometryChange()
            gi.update_shape()

    def create_element_model(self, gscene):
        ele_model = self.app_manager.model.ele_model
        ele_model.elements_from_sequence(self.app_manager.model.seq_model)
        for e in ele_model.elements:
            ge = OpticalElement(e)
            gscene.addItem(ge)

    def create_ray_model(self, gscene, start_surf=1):
        opt_model = self.app_manager.model

        img_dist = abs(opt_model.optical_spec.parax_data[2].img_dist)
        start_offset = 0.05*(gscene.sceneRect().width() + img_dist)

        fov = opt_model.optical_spec.field_of_view
        for fi, f in enumerate(fov.fields):
            rb = RayBundle(opt_model, fi, start_offset)
            gscene.addItem(rb)

    def create_table_view(self, table_model, table_title, close_callback=None):
        # construct the top level widget
        widget = QWidget()
        # construct the top level layout
        layout = QVBoxLayout(widget)

        tableView = QTableView()
        tableView.setAlternatingRowColors(True)

        # Add table to box layout
        layout.addWidget(tableView)

        # set the layout on the widget
        widget.setLayout(layout)

        sub = self.add_subwindow(widget, ModelInfo(self.app_manager.model,
                                                   cmds.update_table_view,
                                                   (tableView,)))
        sub.setWindowTitle(table_title)

        sub.installEventFilter(self)

        tableView.setModel(table_model)

        tableView.setMinimumWidth(tableView.horizontalHeader().length() +
                                  tableView.horizontalHeader().height())
#                                  The following line should work but returns 0
#                                  tableView.verticalHeader().width())

        view_width = tableView.width()
        view_ht = tableView.height()
        orig_x, orig_y = self.initial_window_offset()
        sub.setGeometry(orig_x, orig_y, view_width, view_ht)

        # table data updated successfully
        table_model.update.connect(self.on_data_changed)

        sub.show()

        return tableView

    def eventFilter(self, obj, event):
        if (event.type() == QEvent.Close):
            print('close event received:', obj)
        return False

    def refresh_gui(self):
        self.app_manager.refresh_gui()

    def refresh_app_ui(self):
        dock.update_dock_windows(self)

    def handle_ideal_imager_command(self, iid, command, specsheet):
        ''' link Ideal Imager Dialog buttons to model actions
        iid: ideal imager dialog
        command: text field with the action - same as button label
        specsheet: the input specsheet used to drive the actions
        '''
        if command == 'Apply':
            opt_model = self.app_manager.model
            opt_model.set_from_specsheet(specsheet)
            self.refresh_gui()
        elif command == 'Close':
            for view, info in self.app_manager.view_dict.items():
                if iid == info[0]:
                    view.close()
        elif command == 'Update':
            opt_model = self.app_manager.model
            specsheet = opt_model.specsheet
            firstorder.specsheet_from_parax_data(opt_model, specsheet)
            iid.specsheet_dict[specsheet.conjugate_type] = specsheet
            iid.update_values()
        elif command == 'New':
            opt_model = cmds.create_new_optical_model_from_specsheet(specsheet)
            self.app_manager.set_model(opt_model)
            for view, info in self.app_manager.view_dict.items():
                if iid == info[0]:
                    w = iid
                    mi = info[1]
                    args = (iid, opt_model)
                    new_mi = ModelInfo(model=opt_model, fct=mi.fct,
                                       args=args, kwargs=mi.kwargs)
                    self.app_manager.view_dict[view] = w, new_mi
            self.refresh_gui()
            self.create_lens_table()
            cmds.create_live_layout_view(opt_model, gui_parent=self)
            cmds.create_paraxial_design_view_v2(opt_model, 'ht',
                                                gui_parent=self)
            self.refresh_gui()

    @pyqtSlot(object, int)
    def on_data_changed(self, rootObj, index):
        self.refresh_gui()


def main():
    logging.basicConfig(filename='rayoptics.log',
                        filemode='w',
                        level=logging.INFO)

    qtapp = QApplication(sys.argv)
    qtwnd = MainWindow(qtapp=qtapp)

    qtwnd.show()
    return qtapp.exec_()


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb 26 20:10:33 2018

@author: Mike
"""


import csv
from pathlib import Path


class RGBTable():
    data_list = []

    def __init__(self, filename='sunset_rgb.csv', data_range=[0., 100.]):
        self.data_range = data_range
        if len(self.data_list) == 0:
            path = Path(__file__).resolve().parent
            with open(path / filename, newline='') as f:
                reader = csv.reader(f)
                for row in reader:
                    self.data_list.append(list(map(int, row)))

    def get_color(self, value):
        fract = ((value - self.data_range[0]) /
                 (self.data_range[1] - self.data_range[0]))
        item_index = int(fract*len(self.data_list))
        return self.data_list[item_index]
